#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the "theanolm decode" command.
"""

import gc
import sys
import os
import logging

import numpy
import theano

from theanolm import Network
from theanolm.backend import TextFileType
from theanolm.backend import get_default_device, log_free_mem
from theanolm.scoring import LatticeBatch, LatticeDecoder, RescoredLattice

def add_arguments(parser):
    """Specifies the command line arguments supported by the "theanolm decode"
    command.

    :type parser: argparse.ArgumentParser
    :param parser: a command line argument parser
    """

    argument_group = parser.add_argument_group("files")
    argument_group.add_argument(
        'model_path', metavar='MODEL-FILE', type=str,
        help='the model file that will be used to compute new word scores')
    argument_group.add_argument(
        '--lattices', metavar='FILE', type=str, nargs='*', default=[],
        help='word lattices to be decoded (default stdin, assumed to be '
             'compressed if the name ends in ".gz")')
    argument_group.add_argument(
        '--lattice-list', metavar='FILE', type=TextFileType('r'),
        help='text file containing a list of word lattices to be decoded (one '
             'path per line, the list and the lattice files are assumed to be '
             'compressed if the name ends in ".gz")')
    argument_group.add_argument(
        '--lattice-format', metavar='FORMAT', type=str, default='slf',
        choices=['slf', 'kaldi'],
        help='format of the lattice files, either "slf" (HTK format, default) '
             'or "kaldi" (a Kaldi lattice archive containing text '
             'CompactLattices')
    argument_group.add_argument(
        '--kaldi-vocabulary', metavar='FILE', type=TextFileType('r'),
        default=None,
        help='mapping of words to word IDs in Kaldi lattices (usually '
             'named words.txt)')
    argument_group.add_argument(
        '--output-file', metavar='FILE', type=TextFileType('w'), default='-',
        help='where to write the best paths through the lattices or the '
             'rescored lattice (default stdout, will be compressed if the name '
             'ends in ".gz")')
    argument_group.add_argument(
        '--num-jobs', metavar='J', type=int, default=1,
        help='divide the set of lattice files into J distinct batches, and '
             'process only batch I')
    argument_group.add_argument(
        '--job', metavar='I', type=int, default=0,
        help='the index of the batch that this job should process, between 0 '
             'and J-1')

    argument_group = parser.add_argument_group("decoding")
    argument_group.add_argument(
        '--output', metavar='FORMAT', type=str, default='ref',
        choices=['ref', 'trn', 'full', 'slf', 'kaldi'],
        help='format of the output, one of "ref" (default, utterance ID '
             'followed by words), "trn" (words followed by utterance ID in '
             'parentheses), "full" (utterance ID, acoustic score, language '
             'score, and number of words, followed by words), "slf" (rescored '
             'lattice in HTK format), "kaldi" (rescored lattice in Kaldi '
             'format)')
    argument_group.add_argument(
        '--n-best', metavar='N', type=int, default=1,
        help='print N best paths of each lattice (default 1)')
    argument_group.add_argument(
        '--nnlm-weight', metavar='LAMBDA', type=float, default=1.0,
        help="language model probabilities given by the model read from "
             "MODEL-FILE will be weighted by LAMBDA, when interpolating with "
             "the language model probabilities in the lattice (default is 1.0, "
             "meaning that the LM probabilities in the lattice will be "
             "ignored)")
    argument_group.add_argument(
        '--lm-scale', metavar='LMSCALE', type=float, default=None,
        help="scale language model log probabilities by LMSCALE when computing "
             "the total probability of a path (default is to use the LM scale "
             "specified in the lattice file, or 1.0 if not specified)")
    argument_group.add_argument(
        '--wi-penalty', metavar='WIP', type=float, default=None,
        help="penalize word insertion by adding WIP to the total log "
             "probability as many times as there are words in the path "
             "(without scaling WIP by LMSCALE)")
    argument_group.add_argument(
        '--log-base', metavar='B', type=int, default=None,
        help="convert output log probabilities to base B and WIP from base B "
             "(default is natural logarithm; this does not affect reading "
             "lattices, since they specify their internal log base)")
    argument_group.add_argument(
        '--unk-penalty', metavar='LOGPROB', type=float, default=None,
        help="use constant LOGPROB as <unk> token score (default is to use the "
             "network to predict <unk> probability)")
    argument_group.add_argument(
        '--shortlist', action="store_true",
        help='distribute <unk> token probability among the out-of-shortlist '
             'words according to their unigram frequencies in the training '
             'data')
    argument_group.add_argument(
        '--unk-from-lattice', action="store_true",
        help='use only the probability from the lattice for <unk> tokens')
    argument_group.add_argument(
        '--linear-interpolation', action="store_true",
        help="use linear interpolation of language model probabilities, "
             "instead of (pseudo) log-linear")

    argument_group = parser.add_argument_group("pruning")
    argument_group.add_argument(
        '--max-tokens-per-node', metavar='T', type=int, default=None,
        help="keep only at most T tokens at each node when decoding a lattice "
             "(default is no limit)")
    argument_group.add_argument(
        '--beam', metavar='B', type=float, default=None,
        help="prune tokens whose log probability is at least B smaller than "
             "the log probability of the best token at any given time (default "
             "is no beam pruning)")
    argument_group.add_argument(
        '--recombination-order', metavar='O', type=int, default=None,
        help="keep only the best token, when at least O previous words are "
             "identical (default is to recombine tokens only if the entire "
             "word history matches)")
    argument_group.add_argument(
        '--prune-relative', metavar='R', type=int, default=None,
        help="if set, tighten the beam and the max-tokens-per-node pruning "
             "linearly in the number of tokens in a node; those parameters "
             "will be divided by the number of tokens and multiplied by R")
    argument_group.add_argument(
        '--abs-min-max-tokens', metavar='T', type=float, default=30,
        help="if prune-extra-limit is used, do not tighten max-tokens-per-node "
             "further than this (default is 30)")
    argument_group.add_argument(
        '--abs-min-beam', metavar='B', type=float, default=150,
        help="if prune-extra-limit is used, do not tighten the beam further "
             "than this (default is 150)")

    argument_group = parser.add_argument_group("configuration")
    argument_group.add_argument(
        '--default-device', metavar='DEVICE', type=str, default=None,
        help='when multiple GPUs are present, use DEVICE as default')

    argument_group = parser.add_argument_group("logging and debugging")
    argument_group.add_argument(
        '--log-file', metavar='FILE', type=str, default='-',
        help='path where to write log file (default is standard output)')
    argument_group.add_argument(
        '--log-level', metavar='LEVEL', type=str, default='info',
        help='minimum level of events to log, one of "debug", "info", "warn" '
             '(default "info")')
    argument_group.add_argument(
        '--debug', action="store_true",
        help='enables debugging Theano errors')
    argument_group.add_argument(
        '--profile', action="store_true",
        help='enables profiling Theano functions')

def decode(args):
    """A function that performs the "theanolm decode" command.

    :type args: argparse.Namespace
    :param args: a collection of command line arguments
    """

    log_file = args.log_file
    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        print("Invalid logging level requested:", args.log_level,
              file=sys.stderr)
        sys.exit(1)
    log_format = '%(asctime)s %(funcName)s: %(message)s'
    if args.log_file == '-':
        logging.basicConfig(stream=sys.stdout, format=log_format, level=log_level)
    else:
        logging.basicConfig(filename=log_file, format=log_format, level=log_level)

    if args.debug:
        theano.config.compute_test_value = 'warn'
    else:
        theano.config.compute_test_value = 'off'
    theano.config.profile = args.profile
    theano.config.profile_memory = args.profile

    if (args.lattice_format == 'kaldi') or (args.output == 'kaldi'):
        if args.kaldi_vocabulary is None:
            print("Kaldi lattice vocabulary is not given.", file=sys.stderr)
            sys.exit(1)

    default_device = get_default_device(args.default_device)
    network = Network.from_file(args.model_path,
                                mode=Network.Mode(minibatch=False),
                                default_device=default_device)

    log_scale = 1.0 if args.log_base is None else numpy.log(args.log_base)
    if (args.log_base is not None) and (args.lattice_format == 'kaldi'):
        logging.info("Warning: Kaldi lattice reader doesn't support logarithm "
                     "base conversion.")

    if args.wi_penalty is None:
        wi_penalty = None
    else:
        wi_penalty = args.wi_penalty * log_scale
    decoding_options = {
        'nnlm_weight': args.nnlm_weight,
        'lm_scale': args.lm_scale,
        'wi_penalty': wi_penalty,
        'unk_penalty': args.unk_penalty,
        'use_shortlist': args.shortlist,
        'unk_from_lattice': args.unk_from_lattice,
        'linear_interpolation': args.linear_interpolation,
        'max_tokens_per_node': args.max_tokens_per_node,
        'beam': args.beam,
        'recombination_order': args.recombination_order,
        'prune_relative': args.prune_relative,
        'abs_min_max_tokens': args.abs_min_max_tokens,
        'abs_min_beam': args.abs_min_beam
    }
    logging.debug("DECODING OPTIONS")
    for option_name, option_value in decoding_options.items():
        logging.debug("%s: %s", option_name, str(option_value))

    logging.info("Building word lattice decoder.")
    decoder = LatticeDecoder(network, decoding_options)

    batch = LatticeBatch(args.lattices, args.lattice_list, args.lattice_format,
                         args.kaldi_vocabulary, args.num_jobs, args.job)
    for lattice_number, lattice in enumerate(batch):
        if lattice.utterance_id is None:
            lattice.utterance_id = str(lattice_number)
        logging.info("Utterance `%s´ -- %d of job %d",
                     lattice.utterance_id,
                     lattice_number + 1,
                     args.job)
        log_free_mem()

        final_tokens, recomb_tokens = decoder.decode(lattice)
        if (args.output == "slf") or (args.output == "kaldi"):
            rescored_lattice = RescoredLattice(lattice,
                                               final_tokens,
                                               recomb_tokens,
                                               network.vocabulary)
            rescored_lattice.lm_scale = args.lm_scale
            rescored_lattice.wi_penalty = args.wi_penalty
            if args.output == "slf":
                rescored_lattice.write_slf(args.output_file)
            else:
                assert args.output == "kaldi"
                rescored_lattice.write_kaldi(args.output_file,
                                             batch.kaldi_word_to_id)
        else:
            for token in final_tokens[:min(args.n_best, len(final_tokens))]:
                line = format_token(token,
                                    lattice.utterance_id,
                                    network.vocabulary,
                                    log_scale,
                                    args.output)
                args.output_file.write(line + "\n")
        gc.collect()

def format_token(token, utterance_id, vocabulary, log_scale, output_format):
    """Formats an output line from a token and an utterance ID.

    Reads word IDs from the history list of ``token`` and converts them to words
    using ``vocabulary``. The history may contain also OOV words as text, so any
    ``str`` will be printed literally.

    :type token: Token
    :param token: a token whose history will be formatted

    :type utterance_id: str
    :param utterance_id: utterance ID for full output

    :type vocabulary: Vocabulary
    :param vocabulary: mapping from word IDs to words

    :type log_scale: float
    :param log_scale: divide log probabilities by this number to convert the log
                      base

    :type output_format: str
    :param output_format: which format to write, one of "ref" (utterance ID,
        words), "trn" (words, utterance ID in parentheses), "full" (utterance
        ID, acoustic and LM scores, number of words, words)

    :rtype: str
    :returns: the formatted output line
    """

    words = token.history_words(vocabulary)
    if output_format == 'ref':
        return "{} {}".format(utterance_id, ' '.join(words))
    elif output_format == 'trn':
        return "{} ({})".format(' '.join(words), utterance_id)
    elif output_format == 'full':
        return "{} {} {} {} {} {}".format(
            utterance_id,
            token.ac_logprob / log_scale,
            token.graph_logprob / log_scale,
            token.total_logprob / log_scale,
            len(words),
            ' '.join(words))
    else:
        print("Invalid output format requested:", args.output)
        sys.exit(1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module the implements classes and functions for computing statistics from a
corpus.
"""

import numpy

from theanolm.parsing import utterance_from_line

def compute_word_counts(input_files):
    """Computes word unigram counts using word strings.

    This method does not expect a vocabulary. Start and end of sentence markers
    are not added. Leaves the input files pointing to the beginning of the file.

    :type input_files: list of file or mmap objects
    :param input_files: input text files

    :rtype: dict
    :returns: a mapping from word strings to counts
    """

    result = dict()
    for subset_file in input_files:
        for line in subset_file:
            for word in utterance_from_line(line):
                if word not in result:
                    result[word] = 1
                else:
                    result[word] += 1
        subset_file.seek(0)
    return result

class BigramStatistics(object):
    """Word Unigram and Bigram Counts
    """

    def __init__(self, input_files, vocabulary=None, count_type='int32'):
        """Reads word statistics from corpus file and creates the
        ``unigram_counts`` and ``bigram_counts`` attributes.

        Leaves the input files pointing to the beginning of the file.

        :type input_files: list of file or mmap objects
        :param input_files: input text files

        :type vocabulary: Vocabulary
        :param vocabulary: restrict to these words
        """

        # Don't import scipy unless necessary. This class is only used by
        # wctool.
        from scipy.sparse import dok_matrix

        vocabulary_size = vocabulary.num_words()
        unk_id = vocabulary.word_to_id['<unk>']

        self.unigram_counts = numpy.zeros(vocabulary_size, count_type)
        self.bigram_counts = dok_matrix(
            (vocabulary_size, vocabulary_size), dtype=count_type)

        for subset_file in input_files:
            for line in subset_file:
                sequence = []
                for word in utterance_from_line(line):
                    if word in vocabulary:
                        sequence.append(vocabulary.word_to_id[word])
                    else:
                        sequence.append(unk_id)
                for word_id in sequence:
                    self.unigram_counts[word_id] += 1
                for left_word_id, right_word_id in zip(sequence[:-1],
                                                       sequence[1:]):
                    self.bigram_counts[left_word_id,right_word_id] += 1
            subset_file.seek(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that defines the Parameters class.
"""

import logging
import theano
from theanolm.backend.exceptions import IncompatibleStateError
from theanolm.backend.exceptions import TheanoConfigurationError

class Parameters:
    """Theano Function Parameters

    A dictionary of Theano shared variables. The values can be accessed through
    their path, which also acts as their identifier when they are saved to a
    HDF5 file.
    """

    def __init__(self):
        """Initializes an empty parameter dictionary.
        """

        self._vars = dict()
        self.total_size = 0

    def __getitem__(self, path):
        """Returns a shared variable given parameter path.

        :type path: str
        :param path: parameter path

        :rtype: SharedVariable
        :returns: the corresponding Theano shared variable
        """

        return self._vars[path]

    def add(self, path, value, device=None):
        """Adds a new parameter.

        :type path: str
        :param path: identifier for the shared variable in Theano and its value
                     when stored in a HDF5 file

        :type value: numpy.ndarray
        :param value: initial value for the shared variable

        :type device: str
        :param device: if other than ``None``, the shared variable will be
                       kept in this device
        """

        if path in self._vars:
            raise ValueError("Path `{}´ already in parameters.".format(path))
        if theano.config.device.startswith('gpu') and value.dtype == 'float64':
            raise TheanoConfigurationError(
                'You are using Theano with the old GPU backend ("device=gpu"), '
                'and the parameter {} is float64. This is very inefficient, so '
                'you most likely want to set "floatX=float32".'.format(path))

        if device is None:
            self._vars[path] = theano.shared(value, path)
        else:
            try:
                self._vars[path] = theano.shared(value, path, target=device)
            except TypeError:
                raise RuntimeError(
                    "Unable to create Theano shared variable for parameter {} "
                    "on device {}. If you are using the old backend, you "
                    "cannot assign layers to different GPU devices."
                    .format(path, device))

        logging.debug("     * %s size=%d type=%s device=%s",
                      path, value.size, value.dtype, str(device))
        self.total_size += value.size

    def get_state(self, state):
        """Pulls values from the shared variables into a HDF5 file.

        If there already is a parameter in the file, it will be replaced, so it
        has to have the same number of elements.

        :type state: h5py.File
        :param state: HDF5 file for storing the parameters
        """

        for path, param in self._vars.items():
            if path in state:
                state[path][:] = param.get_value()
            else:
                state.create_dataset(path, data=param.get_value())

    def set_state(self, state):
        """Sets the values of the shared variables.

        Requires that ``state`` contains values for all the parameters.

        :type state: h5py.File
        :param state: HDF5 file that contains the parameters
        """

        for path, param in self._vars.items():
            if path not in state:
                raise IncompatibleStateError(
                    "Parameter `%s´ is missing from state." % path)
            new_value = state[path].value
            param.set_value(new_value)
            if len(new_value.shape) == 0:
                logging.debug("%s <- %s", path, str(new_value))
            else:
                logging.debug("%s <- array%s", path, str(new_value.shape))

    def get_variables(self):
        """Returns a list of the shared variables.

        :rtype: list of strs
        :returns: parameter paths
        """

        return self._vars

    def __str__(self):
        """Returns a string representation of the parameters.

        :rtype: str
        :returns: string representation of the parameters
        """

        return str(self._vars)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the RMSProp optimizer with Nesterov momentum.
"""

import numpy
import theano.tensor as tensor

from theanolm.backend import Parameters
from theanolm.training.basicoptimizer import BasicOptimizer

class RMSPropNesterovOptimizer(BasicOptimizer):
    """RMSProp Variation of Nesterov Momentum Optimization Method

    At the time of writing, RMSProp is an unpublished method. Usually people
    cite slide 29 of Lecture 6 of Geoff Hinton's Coursera class:
    http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf

    The idea is simply to maintain a running average of the squared gradient for
    each parameter, and divide the gradient by the root of the mean squared
    gradient (RMS). This makes RMSProp take steps near 1 whenever the gradient
    is of constant magnitude, and larger steps whenever the local scale of the
    gradient starts to increase.

    RMSProp has been implemented over many optimization methods. This
    implementation is based on the Nesterov Momentum method. We use an
    alternative formulation that requires the gradient to be computed only at
    the current parameter values, described here:
    https://github.com/lisa-lab/pylearn2/pull/136#issuecomment-10381617
    except that we divide the gradient by the RMS gradient:

    rmsprop_{t-1} = -lr * gradient(params_{t-1}) / rms_gradient(params_{t-1})
    v_{t} = mu * v_{t-1} + rmsprop_{t-1}
    params_{t} = params_{t-1} + mu * v_{t} + rmsprop_{t-1}
    """

    def __init__(self, optimization_options, network, *args, **kwargs):
        """Creates an RMSProp momentum optimizer.

        :type optimization_options: dict
        :param optimization_options: a dictionary of optimization options

        :type network: Network
        :param network: the neural network object
        """

        self._params = Parameters()
        for path, param in network.get_variables().items():
            # Initialize mean squared gradient to ones, otherwise the first
            # update will be divided by close to zero.
            self._params.add(path + '_mean_sqr_gradient',
                             numpy.ones_like(param.get_value()))
            self._params.add(path + '_velocity',
                             numpy.zeros_like(param.get_value()))

        # geometric rate for averaging gradients
        if 'gradient_decay_rate' not in optimization_options:
            raise ValueError("Gradient decay rate is not given in training "
                             "options.")
        self._gamma = optimization_options['gradient_decay_rate']

        # momentum
        if 'momentum' not in optimization_options:
            raise ValueError("Momentum is not given in optimization options.")
        self._momentum = optimization_options['momentum']

        super().__init__(optimization_options, network, *args, **kwargs)

    def _get_param_updates(self, alpha):
        """Returns Theano expressions for updating the model parameters and any
        additional parameters required by the optimizer.

        :type alpha: Variable
        :param alpha: a scale to be applied to the model parameter updates

        :rtype: iterable over pairs (shared variable, new expression)
        :returns: expressions how to update the optimizer parameters
        """

        result = []
        deltas = dict()
        for path, gradient in zip(self.network.get_variables(),
                                  self._gradients):
            ms_gradient_old = self._params[path + '_mean_sqr_gradient']
            ms_gradient = \
                self._gamma * ms_gradient_old + \
                (1.0 - self._gamma) * tensor.sqr(gradient)
            result.append((ms_gradient_old, ms_gradient))

            rms_gradient = tensor.sqrt(ms_gradient + self._epsilon)
            deltas[path] = -gradient / rms_gradient
        self._normalize(deltas)

        result = []
        for path, param_old in self.network.get_variables().items():
            delta = deltas[path]
            velocity_old = self._params[path + '_velocity']
            velocity = self._momentum * velocity_old + alpha * delta
            param = param_old + self._momentum * velocity + alpha * delta
            result.append((velocity_old, velocity))
            result.append((param_old, param))
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Modules that implement iterator classes and function related to parsing text.
"""

from theanolm.parsing.linearbatchiterator import LinearBatchIterator
from theanolm.parsing.shufflingbatchiterator import ShufflingBatchIterator
from theanolm.parsing.scoringbatchiterator import ScoringBatchIterator
from theanolm.parsing.functions import utterance_from_line



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import numpy
import theano
from theano import sparse
from theano import tensor
from wordclasses.bigramoptimizer import BigramOptimizer
from wordclasses.functions import byte_size

class TheanoBigramOptimizer(BigramOptimizer):
    """Word Class Optimizer
    """

    def __init__(self, statistics, vocabulary):
        """Computes initial statistics.

        :type statistics: BigramStatistics
        :param statistics: word statistics from the training corpus

        :type vocabulary: theanolm.Vocabulary
        :param vocabulary: words to include in the optimization and initial classes
        """

	# count_nonzero() and any() seem to fail on the sparse matrix.
        if not statistics.unigram_counts.any():
            raise ValueError("Empty word unigram statistics.")
        if statistics.bigram_counts.nnz == 0:
            raise ValueError("Empty word bigram statistics.")

        # Sparse classes in Theano 0.8 support only int32 indices.
        super().__init__(vocabulary, 'int32')

        # Create word counts.
        word_counts = statistics.unigram_counts
        self._word_counts = theano.shared(word_counts, 'word_counts')
        logging.debug("Allocated %s for word counts.",
                      byte_size(word_counts.nbytes))
        ww_counts_csc = statistics.bigram_counts.tocsc()
        self._ww_counts = theano.shared(ww_counts_csc, 'ww_counts_csc')
        logging.debug("Allocated %s for CSC word-word counts.",
                      byte_size(ww_counts_csc.data.nbytes))
        ww_counts_csr = statistics.bigram_counts.tocsr()
        self._ww_counts_csr = theano.shared(ww_counts_csr, 'ww_counts_csr')
        logging.debug("Allocated %s for CSR word-word counts.",
                      byte_size(ww_counts_csr.data.nbytes))

        # Initialize classes.
        word_to_class = numpy.array(vocabulary.word_id_to_class_id)
        self._word_to_class = theano.shared(word_to_class, 'word_to_class')
        logging.debug("Allocated %s for word-to-class mapping.",
                      byte_size(word_to_class.nbytes))

        # Compute class counts from word counts.
        logging.info("Computing class and class/word statistics.")
        class_counts, cc_counts, cw_counts, wc_counts = \
            self._compute_class_statistics(word_counts,
                                           ww_counts_csc,
                                           word_to_class)
        self._class_counts = theano.shared(class_counts, 'class_counts')
        self._cc_counts = theano.shared(cc_counts, 'cc_counts')
        self._cw_counts = theano.shared(cw_counts, 'cw_counts')
        self._wc_counts = theano.shared(wc_counts, 'wc_counts')

        # Create Theano functions.
        self._create_get_word_prob_function()
        self._create_evaluate_function()
        self._create_move_function()
        self._create_log_likelihood_function()
        self._create_class_size_function()

    def get_word_class(self, word_id):
        """Returns the class the given word is currently assigned to.

        :type word_id: int
        :param word_id: ID of the word

        :rtype: int
        :returns: ID of the class the word is assigned to
        """

        return self._word_to_class.get_value()[word_id]

    def _create_get_word_prob_function(self):
        """Creates a Theano function that returns the unigram probability of a
        word within its class.
        """

        word_id = tensor.scalar('word_id', dtype=self._count_type)
        word_id.tag.test_value = 0

        word_count = self._word_counts[word_id]
        class_id = self._word_to_class[word_id]
        class_count = self._class_counts[class_id]
        result = tensor.switch(tensor.neq(class_count, 0),
                               word_count / class_count,
                               0)

        self.get_word_prob = theano.function(
            [word_id],
            result,
            name='get_word_prob')

    def _create_evaluate_function(self):
        """Creates a Theano function that evaluates how much moving a word to
        another class would change the log likelihood.
        """

        word_id = tensor.scalar('word_id', dtype=self._count_type)
        word_id.tag.test_value = 0
        new_class_id = tensor.scalar('new_class_id', dtype=self._count_type)
        new_class_id.tag.test_value = 0

        old_class_id = self._word_to_class[word_id]
        old_class_count = self._class_counts[old_class_id]
        new_class_count = self._class_counts[new_class_id]
        word_count = self._word_counts[word_id]

        # old class
        old_count = old_class_count
        new_count = old_count - word_count
        result = 2 * old_count * tensor.log(old_count)
        result -= 2 * new_count * tensor.log(new_count)

        # new class
        old_count = new_class_count
        new_count = old_count + word_count
        result += 2 * old_count * tensor.log(old_count)
        result -= 2 * new_count * tensor.log(new_count)

        # Iterate over classes other than the old and new class of the word.
        class_ids = tensor.arange(self.num_classes)
        selector = tensor.neq(class_ids, old_class_id) * \
                   tensor.neq(class_ids, new_class_id)
        iter_class_ids = class_ids[selector.nonzero()]

        # old class, class X
        old_counts = self._cc_counts[old_class_id,iter_class_ids]
        new_counts = old_counts - self._wc_counts[word_id,iter_class_ids]
        result -= self._xlogx(old_counts).sum()
        result += self._xlogx(new_counts).sum()

        # new class, class X
        old_counts = self._cc_counts[new_class_id,iter_class_ids]
        new_counts = old_counts + self._wc_counts[word_id,iter_class_ids]
        result -= self._xlogx(old_counts).sum()
        result += self._xlogx(new_counts).sum()

        # class X, old class
        old_counts = self._cc_counts[iter_class_ids,old_class_id]
        new_counts = old_counts - self._cw_counts[iter_class_ids,word_id]
        result -= self._xlogx(old_counts).sum()
        result += self._xlogx(new_counts).sum()

        # class X, new class
        old_counts = self._cc_counts[iter_class_ids,new_class_id]
        new_counts = old_counts + self._cw_counts[iter_class_ids,word_id]
        result -= self._xlogx(old_counts).sum()
        result += self._xlogx(new_counts).sum()

        # old class, new class
        old_count = self._cc_counts[old_class_id,new_class_id]
        new_count = old_count - \
                    self._wc_counts[word_id,new_class_id] + \
                    self._cw_counts[old_class_id,word_id] - \
                    self._ww_counts[word_id,word_id]
        result += self._ll_change(old_count, new_count)

        # new class, old class
        old_count = self._cc_counts[new_class_id,old_class_id]
        new_count = old_count - \
                    self._cw_counts[new_class_id,word_id] + \
                    self._wc_counts[word_id,old_class_id] - \
                    self._ww_counts[word_id,word_id]
        result += self._ll_change(old_count, new_count)

        # old class, old class
        old_count = self._cc_counts[old_class_id,old_class_id]
        new_count = old_count - \
                    self._wc_counts[word_id,old_class_id] - \
                    self._cw_counts[old_class_id,word_id] + \
                    self._ww_counts[word_id,word_id]
        result += self._ll_change(old_count, new_count)

        # new class, new class
        old_count = self._cc_counts[new_class_id,new_class_id]
        new_count = old_count + \
                    self._wc_counts[word_id,new_class_id] + \
                    self._cw_counts[new_class_id,word_id] + \
                    self._ww_counts[word_id,word_id]
        result += self._ll_change(old_count, new_count)

        self._evaluate = theano.function(
            [word_id, new_class_id],
            result,
            name='evaluate')

    def _create_move_function(self):
        """Creates a Theano function that moves a word to another class.

        tensor.inc_subtensor actually works like numpy.add.at, so we can use it
        to add the count as many times as the word occurs in a class.
        """

        updates = []
        word_id = tensor.scalar('word_id', dtype=self._count_type)
        word_id.tag.test_value = 0
        new_class_id = tensor.scalar('new_class_id', dtype=self._count_type)
        new_class_id.tag.test_value = 0
        old_class_id = self._word_to_class[word_id]

        # word
        word_count = self._word_counts[word_id]
        c_counts = self._class_counts
        c_counts = tensor.inc_subtensor(c_counts[old_class_id], -word_count)
        c_counts = tensor.inc_subtensor(c_counts[new_class_id], word_count)
        updates.append((self._class_counts, c_counts))

        # word, word X
        data, indices, indptr, _ = sparse.csm_properties(self._ww_counts_csr)
        right_word_ids = indices[indptr[word_id]:indptr[word_id + 1]]
        counts = data[indptr[word_id]:indptr[word_id + 1]]
        selector = tensor.neq(right_word_ids, word_id).nonzero()
        right_word_ids = right_word_ids[selector]
        counts = counts[selector]

        cw_counts = self._cw_counts
        cw_counts = tensor.inc_subtensor(cw_counts[old_class_id,right_word_ids], -counts)
        cw_counts = tensor.inc_subtensor(cw_counts[new_class_id,right_word_ids], counts)
        right_class_ids = self._word_to_class[right_word_ids]
        cc_counts = self._cc_counts
        cc_counts = tensor.inc_subtensor(cc_counts[old_class_id,right_class_ids], -counts)
        cc_counts = tensor.inc_subtensor(cc_counts[new_class_id,right_class_ids], counts)

        # word X, word
        data, indices, indptr, _ = sparse.csm_properties(self._ww_counts)
        left_word_ids = indices[indptr[word_id]:indptr[word_id + 1]]
        counts = data[indptr[word_id]:indptr[word_id + 1]]
        selector = tensor.neq(left_word_ids, word_id).nonzero()
        left_word_ids = left_word_ids[selector]
        counts = counts[selector]

        wc_counts = self._wc_counts
        wc_counts = tensor.inc_subtensor(wc_counts[left_word_ids,old_class_id], -counts)
        wc_counts = tensor.inc_subtensor(wc_counts[left_word_ids,new_class_id], counts)
        left_class_ids = self._word_to_class[left_word_ids]
        cc_counts = tensor.inc_subtensor(cc_counts[left_class_ids,old_class_id], -counts)
        cc_counts = tensor.inc_subtensor(cc_counts[left_class_ids,new_class_id], counts)

        # word, word
        count = self._ww_counts[word_id,word_id]
        cc_counts = tensor.inc_subtensor(cc_counts[old_class_id,old_class_id], -count)
        cc_counts = tensor.inc_subtensor(cc_counts[new_class_id,new_class_id], count)
        cw_counts = tensor.inc_subtensor(cw_counts[old_class_id,word_id], -count)
        cw_counts = tensor.inc_subtensor(cw_counts[new_class_id,word_id], count)
        wc_counts = tensor.inc_subtensor(wc_counts[word_id,old_class_id], -count)
        wc_counts = tensor.inc_subtensor(wc_counts[word_id,new_class_id], count)
        updates.append((self._cc_counts, cc_counts))
        updates.append((self._cw_counts, cw_counts))
        updates.append((self._wc_counts, wc_counts))

        w_to_c = self._word_to_class
        w_to_c = tensor.set_subtensor(w_to_c[word_id], new_class_id)
        updates.append((self._word_to_class, w_to_c))

        self._move = theano.function(
            [word_id, new_class_id],
            [],
            updates=updates,
            name='move')

    def _create_log_likelihood_function(self):
        """Creates a Theano function that computes the log likelihood that a
        bigram model would give to the corpus.
        """

        result = self._xlogx(self._cc_counts).sum() + \
                 self._xlogx(self._word_counts).sum() - \
                 2 * self._xlogx(self._class_counts).sum()

        self.log_likelihood = theano.function(
            [],
            result,
            name='log_likelihood')

    def _create_class_size_function(self):
        """Creates a function that calculates the number of words in a class.

        :type class_id: int
        :param class_id: ID of a class

        :rtype: int
        :returns: number of words in the class
        """

        class_id = tensor.scalar('class_id', dtype=self._count_type)
        class_id.tag.test_value = 0

        result = tensor.eq(self._word_to_class, class_id).sum()

        self._class_size = theano.function(
            [class_id],
            result,
            name='class_size')

    @staticmethod
    def _ll_change(old_count, new_count):
        """A helper function that computes the log likelihood change when a
        count changes.
        """

        result = 0
        if old_count != 0:
            result -= old_count * tensor.log(old_count)
        if new_count != 0:
            result += new_count * tensor.log(new_count)
        return result

    @staticmethod
    def _xlogx(x):
        """A helper function that computes ``x * log(x)``, where ``x`` is a
        vector that may contain zeros.
        """

        return tensor.switch(tensor.neq(x, 0), tensor.log(x), 0) * x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the projection layer.
"""

import theano.tensor as tensor

from theanolm.network.basiclayer import BasicLayer

class ProjectionLayer(BasicLayer):
    """Projection Layer

    Projection layer supports dividing the weight to multiple devices. The
    second dimension of the projection matrix (output space) is split to equal
    parts. After projecting on each part, the output will be concatenated.
    """

    def __init__(self, *args, **kwargs):
        """Initializes the parameters used by this layer.
        """

        super().__init__(*args, **kwargs)

        # Initialize the parameters.
        input_size = sum(x.output_size for x in self._input_layers)
        output_size = self.output_size
        self._init_weight('W', (input_size, output_size), scale=0.01)

        self.output = None

    def create_structure(self):
        """Creates the symbolic graph of this layer.

        The input is always 2-dimensional: the first dimension is the time step
        (index of word in a sequence) and the second dimension are the
        sequences.

        Sets self.output to a symbolic matrix that describes the output of this
        layer.
        """

        assert len(self._input_layers) == 1
        layer_input = self._input_layers[0].output
        num_time_steps = layer_input.shape[0]
        num_sequences = layer_input.shape[1]

        outputs = []
        for device in self._devices:
            # Indexing the word_projection matrix with a word ID returns the
            # self.output_size dimensional projection. Note that indexing the
            # matrix with a vector of all the word IDs gives a concatenation of
            # those projections.
            device_output = self._get_param('W', device)[layer_input.flatten()]
            device_output = device_output.reshape([num_time_steps,
                                                   num_sequences,
                                                   -1])
            outputs.append(device_output)
        self.output = tensor.concatenate(outputs, axis=2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the addition layer.
"""

import theano.tensor as tensor

from theanolm.network.basiclayer import BasicLayer

class AdditionLayer(BasicLayer):
    """Addition Layer

    A layer that produces an elementwise sum of its inputs. All inputs have to
    be equal size.
    """

    def __init__(self, layer_options, *args, **kwargs):
        """If the input dimensionality does not change, this layer has no
        parameters. A weight matrix for matching the dimensions is created for
        every input that has different dimensionality than the output.

        If the output size is not specified, the size of the first input will be
        assumed.

        :type layer_options: dict
        :param layer_options: dictionary of layer options
        """

        if 'size' not in layer_options:
            layer_options['size'] = layer_options['input_layers'][0].output_size

        super().__init__(layer_options, *args, **kwargs)

        for input_index, input_layer in enumerate(self._input_layers):
            input_size = input_layer.output_size
            if input_size != self.output_size:
                param_name = 'input{}/W'.format(input_index)
                self._init_weight(param_name, (input_size, self.output_size),
                                  scale=0.01)

        self.output = None

    def create_structure(self):
        """Creates the symbolic graph of this layer.

        Sets self.output to a symbolic matrix that describes the output of this
        layer. If the inputs are the same size as the output, the output will be
        the elementwise sum of the inputs. If needed, the inputs will be
        projected to the same size.
        """

        for input_index, input_layer in enumerate(self._input_layers):
            input_size = input_layer.output_size
            if input_size == self.output_size:
                input_matrix = input_layer.output
            else:
                input_matrix = self._tensor_preact(input_layer.output,
                                                   'input{}'.format(input_index),
                                                   use_bias=False)

            if self.output is None:
                self.output = input_matrix
            else:
                self.output = tensor.add(self.output, input_matrix)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Functions related to creating Theano matrices.
"""

import numpy
import theano

def test_value(size, high):
    """Creates a matrix of random numbers that can be used as a test value for a
    parameter to enable debugging Theano errors.

    The type of ``high`` defines the type of the returned array. For integers,
    the range does not include the maximum value. If ``high`` is a boolean,
    returns an int8 array, as Theano uses int8 to represent a boolean.

    :type size: int or tuple of ints
    :param size: dimensions of the matrix

    :type high: int, float, or bool
    :param high: maximum value for the generated random numbers

    :rtype: numpy.ndarray
    :returns: a matrix or vector containing the generated values
    """

    if isinstance(high, bool):
        return numpy.random.randint(0, int(high), size=size).astype('int8')
    elif isinstance(high, (int, numpy.int32, numpy.int64)):
        return numpy.random.randint(0, high, size=size).astype('int64')
    elif isinstance(high, (float, numpy.float32, numpy.float64)):
        return high * numpy.random.rand(*size).astype(theano.config.floatX)
    else:
        raise TypeError("High value should be int, float, or bool.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements custom exception classes.
"""

class InputError(Exception):
    """Exception raised for errors in the input.
    """
    pass

class NumberError(Exception):
    """Exception raised when one of the parameter gets NaN value.
    """
    pass

class IncompatibleStateError(Exception):
    """Exception raised when attempting to load a state that is incompatible
    with the neural network structure.
    """
    pass

class TheanoConfigurationError(Exception):
    """Exception raised when Theano has not been properly configured.
    """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the ADADELTA optimizer.
"""

import numpy
import theano.tensor as tensor

from theanolm.backend import Parameters
from theanolm.training.basicoptimizer import BasicOptimizer

class AdadeltaOptimizer(BasicOptimizer):
    """ADADELTA Optimization Method

    ADADELTA optimization method has been derived from AdaGrad. AdaGrad
    accumulates the sum of squared gradients over all time, which is used to
    scale the learning rate smaller and smaller. ADADELTA uses an exponentially
    decaying average of the squared gradients.

    This implementation scales the parameter updates by the learning rate
    hyperparameter. The original paper does not include such scaling,
    corresponding to learning rate 1.

    M. D. Zeiler (2012)
    ADADELTA: An adaptive learning rate method
    http://arxiv.org/abs/1212.5701
    """

    def __init__(self, optimization_options, network, *args, **kwargs):
        """Creates an Adadelta optimizer.

        :type optimization_options: dict
        :param optimization_options: a dictionary of optimization options

        :type network: Network
        :param network: the neural network object
        """

        self._params = Parameters()
        for path, param in network.get_variables().items():
            self._params.add(path + '_mean_sqr_gradient',
                             numpy.zeros_like(param.get_value()))
            self._params.add(path + '_mean_sqr_delta',
                             numpy.zeros_like(param.get_value()))

        # geometric rate for averaging gradients
        if 'gradient_decay_rate' not in optimization_options:
            raise ValueError("Gradient decay rate is not given in optimization "
                             "options.")
        self._gamma = optimization_options['gradient_decay_rate']

        super().__init__(optimization_options, network, *args, **kwargs)

    def _get_param_updates(self, alpha):
        """Returns Theano expressions for updating the model parameters and any
        additional parameters required by the optimizer.

        :type alpha: Variable
        :param alpha: a scale to be applied to the model parameter updates

        :rtype: iterable over pairs (shared variable, new expression)
        :returns: expressions how to update the optimizer parameters
        """

        result = []
        deltas = dict()
        for path, gradient in zip(self.network.get_variables(),
                                  self._gradients):
            ms_gradient_old = self._params[path + '_mean_sqr_gradient']
            ms_gradient = self._gamma * ms_gradient_old + \
                          (1.0 - self._gamma) * tensor.sqr(gradient)
            result.append((ms_gradient_old, ms_gradient))

            # New delta is expressed recursively using the current RMS gradient
            # and the old delta.
            ms_delta_old = self._params[path + '_mean_sqr_delta']
            rms_delta_old = tensor.sqrt(ms_delta_old + self._epsilon)
            rms_gradient = tensor.sqrt(ms_gradient + self._epsilon)
            deltas[path] = -gradient * rms_delta_old / rms_gradient
        self._normalize(deltas)

        for path, param_old in self.network.get_variables().items():
            delta = deltas[path]
            ms_delta_old = self._params[path + '_mean_sqr_delta']
            ms_delta = self._gamma * ms_delta_old + \
                       (1.0 - self._gamma) * tensor.sqr(delta)
            result.append((ms_delta_old, ms_delta))
            result.append((param_old, param_old + alpha * delta))
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the AdaGrad optimizer.
"""

import numpy
import theano.tensor as tensor

from theanolm.backend import Parameters
from theanolm.training.basicoptimizer import BasicOptimizer

class AdaGradOptimizer(BasicOptimizer):
    """AdaGrad Optimization Method

    AdaGrad is a simple extension of Stochastic Gradient Descent that adapts the
    step size for each component, based on how frequently each component occurs
    in the gradients. At each update, the learning rate is divided by the root
    of the sum of squared gradients. (Actually, in this simpler form of the
    algorithm, the squared gradient is used to approximate the outer product of
    the gradient vector by itself.)

    J. Duchi, E. Hazan, Y. Singer (2011)
    Adaptive Subgradient Methods for Online Learning and Stochastic Optimization
    Journal of Machine Learning Research 12: 2121-2159

    Note: When using a learning rate decreasing schedule, perhaps a running
    average of the historical gradients would be better than a sum.
    """

    def __init__(self, optimization_options, network, *args, **kwargs):
        """Creates an AdaGrad optimizer.

        :type optimization_options: dict
        :param optimization_options: a dictionary of optimization options

        :type network: Network
        :param network: the neural network object
        """

        self._params = Parameters()
        for path, param in network.get_variables().items():
            self._params.add(path + '_sum_sqr_gradient',
                             numpy.zeros_like(param.get_value()))

        super().__init__(optimization_options, network, *args, **kwargs)

    def _get_param_updates(self, alpha):
        """Returns Theano expressions for updating the model parameters and any
        additional parameters required by the optimizer.

        :type alpha: Variable
        :param alpha: a scale to be applied to the model parameter updates

        :rtype: iterable over pairs (shared variable, new expression)
        :returns: expressions how to update the optimizer parameters
        """

        result = []
        deltas = dict()
        for path, gradient in zip(self.network.get_variables(),
                                  self._gradients):
            ss_gradient_old = self._params[path + '_sum_sqr_gradient']
            ss_gradient = ss_gradient_old + tensor.sqr(gradient)
            result.append((ss_gradient_old, ss_gradient))

            rss_gradient = tensor.sqrt(ss_gradient + self._epsilon)
            deltas[path] = -gradient / rss_gradient
        self._normalize(deltas)

        for path, param_old in self.network.get_variables().items():
            delta = deltas[path]
            result.append((param_old, param_old + alpha * delta))
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the WordClass class.
"""

from collections import OrderedDict

import numpy

class WordClass(object):
    """Collection of Words and Their Membership Probabilities

    A word class contains one or more words and their probabilities within
    the class. When a class-based model is not wanted, word classes will be
    created with exactly one word per class.

    The class does not enforce the membership probabilities to sum up to
    one. The user has to call ``normalize_probs()`` after creating the
    class.
    """

    def __init__(self, class_id, word_id, prob):
        """Initializes the class with one word with given probability.

        :type class_id: int
        :param class_id: ID for the class

        :type word_id: int
        :param word_id: ID of the initial word in the class

        :type prob: float
        :param prob: the membership probability of the word
        """

        self.id = class_id
        self._probs = OrderedDict({word_id: prob})

    def add(self, word_id, prob):
        """Adds a word to the class with given probability.

        The membership probabilities are not guaranteed to be normalized.

        :type word_id: int
        :param word_id: ID of the word to add to the class

        :type prob: float
        :param prob: the membership probability of the new word
        """

        self._probs[word_id] = prob

    def get_prob(self, word_id):
        """Returns the class membership probability of a word.

        :type word_id: int
        :param word_id: a word ID that belongs to this class

        :rtype: float
        :returns: the class membership probability of the word
        """

        return self._probs[word_id]

    def set_prob(self, word_id, prob):
        """Sets the class membership probability of a word. The word will be
        added to this class, if it doesn't belong to it already.

        :type word_id: int
        :param word_id: a word ID

        :type prob: float
        :param prob: class membership probability for the word
        """

        self._probs[word_id] = prob

    def normalize_probs(self):
        """Normalizes the class membership probabilities to sum to one.
        """

        prob_sum = sum(self._probs.values())
        for word_id in self._probs:
            self._probs[word_id] /= prob_sum

    def sample(self):
        """Samples a word from the membership probability distribution.

        :rtype: int
        :returns: a random word ID from this class
        """

        word_ids = list(self._probs.keys())
        probs = list(self._probs.values())
        sample_distribution = numpy.random.multinomial(1, probs)
        indices = numpy.flatnonzero(sample_distribution)
        assert len(indices) == 1
        return word_ids[indices[0]]

    def __len__(self):
        """Returns the number of words in this class.

        :rtype: int
        :returns: the number of words in this class
        """

        return len(self._probs)

    def __iter__(self):
        """A generator for iterating through the words in this class.

        :rtype: generator for (int, float)
        :returns: generates a tuple containing a word ID and class
                  membership probability
        """

        for word_id, prob in self._probs.items():
            yield word_id, prob

    def __eq__(self, other):
        """Tests if another word class is exactly the same.

        Two word classes are considered the same if the same word IDs have
        the same probabilities within a tolerance.

        :type other: WordClass
        :param other: another word class

        :rtype: bool
        :returns: True if the classes are the same, False otherwise
        """

        if not isinstance(other, self.__class__):
            return False

        if self.id != other.id:
            return False

        if len(self) != len(other):
            return False

        for word_id, prob in self:
            if not numpy.isclose(prob, other._probs[word_id]):
                return False

        return True

    def __ne__(self, other):
        """Tests if another word class is different.

        Two word classes are considered the same if the same word IDs have
        the same probabilities within a tolerance.

        :type other: WordClass
        :param other: another word class

        :rtype: bool
        :returns: False if the classes are the same, True otherwise
        """

        return not self.__eq__(other)

    def __str__(self):
        """Writes the class members in a string.

        :rtype: str
        :returns: a string showing the class members and their
                  probabilities.
        """

        return '{ ' + \
               ', '.join(str(word_id) + ': ' + str(round(prob, 4))
                         for word_id, prob in self) + \
               ' }'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the Lattice class used by word lattice decoders.
"""

import logging

from theanolm.backend import InputError
from theanolm.scoring.lattice import Lattice

class NodeNotFoundError(Exception):
    pass

def _follow_word_from_node(node, word):
    """Follows the link with given word label from given node.

    If there is a link from ``node`` with the label ``word``, returns the end
    node and the log probabilities and transition IDs of the link. If there are
    null links in between, returns the sum of the log probabilities and the
    concatenation of the transition IDs.

    :type node: Lattice.Node
    :param node: node where to start searching

    :type word: str
    :param word: word to search for

    :rtype: tuple of (Lattice.Node, float, float, str)
    :returns: the end node of the link with the word label (or ``None`` if the
              word is not found), and the total acoustic log probability, LM log
              probability, and transition IDs of the path to the word
    """

    if word not in node.word_to_link:
        return (None, None, None, None)

    link = node.word_to_link[word]
    if link.word is not None:
        return (link.end_node,
                link.ac_logprob if link.ac_logprob is not None else 0.0,
                link.lm_logprob if link.lm_logprob is not None else 0.0,
                link.transitions if link.transitions is not None else "")

    end_node, ac_logprob, lm_logprob, transitions = \
        _follow_word_from_node(link.end_node, word)
    if end_node is None:
        return (None, None, None, None)
    else:
        if link.ac_logprob is not None:
            ac_logprob += link.ac_logprob
        if link.lm_logprob is not None:
            lm_logprob += link.lm_logprob
        if link.transitions is not None:
            transitions += link.transitions
        return (end_node, ac_logprob, lm_logprob, transitions)

class RescoredLattice(Lattice):
    """Rescored Lattice

    Word lattice that is created from the tokens after decoding another lattice.
    The LM probabilities in the lattice are replaced with interpolated
    probabilities from the TheanoLM model and those in the original lattice.
    """

    def __init__(self, original_lattice, final_tokens, recomb_tokens,
                 vocabulary):
        """Constructs a lattice from the paths visited when decoding a lattice.

        :type original_lattice: Lattice
        :param original_lattice: the original lattice that was decoded; the new
            lattice will be based on the links found in this lattice

        :type final_tokens: list of LatticeDecoder.Tokens
        :param final_tokens: the tokens that reached the final node in the
            decoder; the resulting lattice will always contain these paths

        :type recomb_tokens: list of tuples
        :param recomb_tokens: tokens that were dropped during recombination;
            each tuple contains the token, and the word IDs and NNLM log
            probability of the token that was kept when dropping the token

        :type vocabulary: Vocabulary
        :param vocabulary: vocabulary for mapping the word IDs in the tokens to
            the words in the lattices

        :rtype: Lattice
        :returns: the created lattice
        """

        def follow_word_ids(word_ids, create=True):
            """Follows a path from the initial node with given word IDs.

            :type word_ids: list of ints
            :param word_ids: IDs of the words to be found on the path

            :type create: bool
            :param create: if ``True``, creates new nodes if necessary

            :rtype: Lattice.Node
            :returns: the last node of the path
            """

            words = [vocabulary.id_to_word[id] if isinstance(id, int) else id
                     for id in word_ids[1:]]
            return self._follow_words(words, original_lattice, create)

        super().__init__()

        self.utterance_id = original_lattice.utterance_id

        self.initial_node = self.Node(len(self.nodes))
        self.initial_node.word_to_link = dict()
        self.nodes.append(self.initial_node)
        final_node = self.Node(None)
        final_node.final = True
        final_node.word_to_link = dict()

        self._add_word_maps(original_lattice.nodes)

        # Create all the paths that correspond to the final tokens.
        for token in final_tokens:
            node = follow_word_ids(token.history)
            # If the lattice is not determinized, it may happen that we have two
            # tokens with the same word sequence.
            if any(link.end_node.final for link in node.out_links):
                continue
            # For the final link we set the NNLM log probability of the entire
            # path. The created lattice is a tree, so we can equivalently set
            # the entire path probability in the final link.
            link = self.Link(node, final_node, word=None,
                             ac_logprob=0,
                             lm_logprob=token.nn_lm_logprob,
                             transitions="")
            node.out_links.append(link)
            self.links.append(link)

        # Add recombined tokens.
        oov_words = set()
        for token, new_history, nn_lm_logprob in reversed(recomb_tokens):
            assert new_history[-1] == token.history[-1]
            word_id = token.history[-1]
            if isinstance(word_id, int):
                word = vocabulary.id_to_word[word_id]
            else:
                word = word_id
                oov_words.add(word)

            # Find the incoming link that corresponds to the token that was kept
            # during recombination.
            try:
                recomb_from_node = follow_word_ids(new_history[:-1], False)
            except NodeNotFoundError:
                continue
            # Our new lattice doesn't contain null links, so word_to_link maps
            # never skip nodes.
            if word not in recomb_from_node.word_to_link:
                continue
            recomb_link = recomb_from_node.word_to_link[word]

            # Add a link from the previous word in the token history to the node
            # that was kept during recombination. The difference in LM log
            # probability can be computed from the token (path) NNLM log
            # probabilities.
            from_node = follow_word_ids(token.history[:-1])
            lm_logprob_diff = token.nn_lm_logprob - nn_lm_logprob
            new_link = self.Link(from_node, recomb_link.end_node, word,
                                 recomb_link.ac_logprob,
                                 recomb_link.lm_logprob + lm_logprob_diff,
                                 recomb_link.transitions)
            from_node.out_links.append(new_link)
            # Tokens never contain null words, so we can be sure that
            # word_to_link maps in our new lattice never skip nodes.
            assert word is not None
            from_node.word_to_link[word] = new_link
            self.links.append(new_link)

        if oov_words:
            logging.debug("Out-of-vocabulary words in lattice: %s",
                          ', '.join(oov_words))

        final_node.id = len(self.nodes)
        self.nodes.append(final_node)

    def _add_word_maps(self, nodes):
        """Adds mapping from words to outgoing links on each node.

        For each node, finds the words that follow the node. If an outgoing link
        is a null link, follows the link recursively. Then adds the attribute
        ``word_to_link`` to the node that provides a mapping from words to the
        links that have to be followed to reach that word. This is done to avoid
        doing the search every time when constructing the rescored lattice.

        :type nodes: list of Lattice.Nodes
        :param nodes: add a ``word_to_link`` map to these nodes
        """

        def next_words(link):
            if link.word is None:
                return [word
                        for next_link in link.end_node.out_links
                        for word in next_words(next_link)]
            else:
                return [link.word]

        determinized = True
        for node in nodes:
            node.word_to_link = {word: link
                                 for link in node.out_links
                                 for word in next_words(link)}
            if len(node.word_to_link) != len(node.out_links):
                determinized = False
        if not determinized:
            logging.warning(
                "The original word lattice is not determinized, meaning that "
                "there are multiple links from the same node with the same "
                "word label, or the lattice contains null links (epsilon "
                "arcs). Lattice rescoring may not work properly.")

    def _follow_words(self, words, original_lattice, create=True):
        """Ensures that a path exists, creating new nodes if necessary, and
        returns the last node of the path.

        The new lattice will be created by repeatedly calling this function. It
        starts from the beginning of the lattice and follows the nodes that
        correspond to the given words. When a word is not found from the
        outgoing links of the node, the tail will be created. Thus the created
        lattice is always a tree.

        :type words: list of strs
        :param words: words of the path

        :type original_lattice: Lattice
        :param original_lattice: the original lattice where the probabilities of
            the links will be read from

        :type create: bool
        :param create: if set to ``False``, won't create any nodes but raises a
            ``NodeNotFoundError`` if the path doesn't exist

        :rtype: Lattice.Node
        :returns: the last node of the path
        """

        node = self.initial_node
        orig_node = original_lattice.initial_node

        for word in words:
            # The original lattice may contain null links that we have to skip.
            orig_end_node, ac_logprob, lm_logprob, transitions = \
                _follow_word_from_node(orig_node, word)
            if orig_end_node is None:
                continue

            # Our new lattice doesn't contain null links, so word_to_link maps
            # never skip nodes.
            if word in node.word_to_link:
                link = node.word_to_link[word]
            else:
                if not create:
                    raise NodeNotFoundError
                end_node_id = len(self.nodes)
                end_node = self.Node(end_node_id)
                end_node.word_to_link = dict()
                self.nodes.append(end_node)
                link = self.Link(node, end_node, word,
                                 ac_logprob, lm_logprob, transitions)
                node.out_links.append(link)
                node.word_to_link[word] = link
                self.links.append(link)

            node = link.end_node
            orig_node = orig_end_node

        return node



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the hierarchical softmax layer.
"""

import logging

import numpy
import theano.tensor as tensor

from theanolm.network.basiclayer import BasicLayer

class HSoftmaxLayer(BasicLayer):
    """Hierarchical Softmax Output Layer

    The output layer is a simple softmax layer that outputs the word
    probabilities.
    """

    def __init__(self, *args, **kwargs):
        """Initializes the parameters used by this layer.
        """

        super().__init__(*args, **kwargs)

        if len(self._devices) > 1:
            raise ValueError("Hierarchical softmax layer does not support "
                             "multiple devices.")

        # Factorize the output size into two levels of equal or almost equal
        # size.
        output_size = self.output_size
        level1_size = numpy.int(numpy.ceil(numpy.sqrt(output_size)))
        level2_size = numpy.int(numpy.ceil(output_size / level1_size))
        assert level1_size * level2_size >= output_size
        assert (level1_size - 1) * level2_size < output_size
        assert level1_size * (level2_size - 1) < output_size
        assert level1_size == level2_size or level1_size == level2_size + 1
        self._level1_size = level1_size
        self._level2_size = level2_size
        logging.debug("  level1_size=%d level2_size=%d",
                      self._level1_size,
                      self._level2_size)

        # Create the parameters. Weight matrix and bias for concatenated input
        # and the second level of the hierarchy.
        input_size = sum(x.output_size for x in self._input_layers)
        self._init_weight('input/W', (input_size, level1_size), scale=0.01)
        self._init_bias('input/b', level1_size)
        self._init_weight('level1/W',
                          (level1_size, input_size, level2_size),
                          scale=0.01)
        self._init_bias('level1/b', (level1_size, level2_size))

        self.output = None
        self.output_probs = None
        self.target_probs = None

    def create_structure(self):
        """Creates the symbolic graph of this layer.

        The input is always 3-dimensional: the first dimension is the time step,
        the second dimension are the sequences, and the third dimension is the
        word projection. When generating text, there's just one sequence and one
        time step in the input.

        Sets self.output to a symbolic matrix that describes the output of this
        layer.
        """

        layer_input = tensor.concatenate([x.output for x in self._input_layers],
                                         axis=2)

        num_time_steps = layer_input.shape[0]
        num_sequences = layer_input.shape[1]
        minibatch_size = num_time_steps * num_sequences
        input_size = layer_input.shape[2]

        input_weight = self._params[self._param_path('input/W')]
        input_bias = self._params[self._param_path('input/b')]
        level1_weight = self._params[self._param_path('level1/W')]
        level1_bias = self._params[self._param_path('level1/b')]

        # First create the output for the whole probability distribution.
        # Combine the first two dimensions so that softmax is taken
        # independently for each location, over the output classes.
        probs = tensor.nnet.h_softmax(layer_input.reshape([minibatch_size,
                                                           input_size]),
                                      minibatch_size,
                                      self.output_size,
                                      self._level1_size,
                                      self._level2_size,
                                      input_weight,
                                      input_bias,
                                      level1_weight,
                                      level1_bias)
        self.output_probs = probs.reshape([num_time_steps,
                                           num_sequences,
                                           self.output_size])

        # Next create the output for target classes. It can only be used when
        # self._network.target_class_ids is given to the function.
        target_class_ids = self._network.target_class_ids.flatten()
        probs = tensor.nnet.h_softmax(layer_input.reshape([minibatch_size,
                                                           input_size]),
                                      minibatch_size,
                                      self.output_size,
                                      self._level1_size,
                                      self._level2_size,
                                      input_weight,
                                      input_bias,
                                      level1_weight,
                                      level1_bias,
                                      target_class_ids)
        self.target_probs = probs.reshape([num_time_steps,
                                           num_sequences])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that provides functions for operating with probabilities, and
defines logprob_type.
"""

from decimal import Decimal, getcontext

import numpy
import theano

logprob_type = numpy.dtype(theano.config.floatX).type

def interpolate_linear(logprob1, logprob2, weight1):
    """Performs linear interpolation of two probabilities.

    If a probability is so small that it will be rounded to zero, uses the
    decimal library. If a weight is zero, the corresponding log probability will
    be ignored. Otherwise if the log probability was ``-inf``, multiplication
    would result in a ``nan``.

    :type logprob1: logprob_type
    :param logprob1: logarithm of first input probability

    :type logprob2: logprob_type
    :param logprob2: logarithm of second input probability

    :type weight1: logprob_type
    :param weight1: interpolation weight for the first probability

    :rtype: logprob_type
    :returns: logarithm of the weighted sum of the input probabilities
    """

    prob1 = numpy.exp(numpy.float64(logprob1))
    prob2 = numpy.exp(numpy.float64(logprob2))
    if (prob1 > 0) and (prob2 > 0):
        weight2 = 1.0 - weight1
        prob = weight1 * prob1
        prob += weight2 * prob2
        return logprob_type(numpy.log(prob))
    else:
        # An exp() resulted in an underflow (or -inf logprob). Use the decimal
        # library.
        getcontext().prec = 16
        d_weight1 = Decimal(float(weight1))
        d_weight2 = Decimal(1.0) - d_weight1
        d_logprob1 = Decimal(float(logprob1))
        d_logprob2 = Decimal(float(logprob2))
        d_zero = Decimal(0.0)
        d_prob = d_zero
        if d_weight1 != d_zero:
            d_prob += d_weight1 * d_logprob1.exp()
        if d_weight2 != d_zero:
            d_prob += d_weight2 * d_logprob2.exp()
        return logprob_type(d_prob.ln())

def interpolate_loglinear(logprob1, logprob2, prior1, prior2):
    """Performs log-linear interpolation of two probabilities.

    This is not real log-linear interpolation, because we don't normalize the
    result. Thus the result is not a real probability.

    If a prior is zero, the corresponding log probability will be ignored.
    Otherwise if the log probability was ``-inf``, multiplication would result
    in a ``nan``.

    :type logprob1: logprob_type
    :param logprob1: first input log probability

    :type logprob2: logprob_type
    :param logprob2: second input log probability

    :type prior1: logprob_type
    :param prior1: weight for the first log probability

    :type prior2: logprob_type
    :param prior2: weight for the second log probability

    :rtype: logprob_type
    :returns: weighted sum of the input log probabilities
    """

    result = 0
    if prior1 != 0:
        result += prior1 * logprob1
    if prior2 != 0:
        result += prior2 * logprob2
    assert not numpy.isnan(result)
    return result



# -*- coding: utf-8 -*-
"""Executable that launches one of the theanolm subcommands.
"""

from traceback import format_tb
import logging
import sys
import argparse

from theanolm.backend import NumberError, TheanoConfigurationError
from theanolm.backend import IncompatibleStateError, InputError
from theanolm.commands import train, score, decode, sample, version

def _get_message(e):
    if hasattr(e, 'args') and len(e.args) > 0 and isinstance(e.args[0], bytes):
        return e.args[0].decode('utf-8')
    return str(e)

# These exception classes are not needed, if gpuarray is not used.
try:
    from pygpu.gpuarray import GpuArrayException
    from theano.gpuarray.type import ContextNotDefined
except ImportError:
    class GpuArrayException(Exception):
        """Dummy GpuArrayException Exception
        """
        pass
    class ContextNotDefined(Exception):
        """Dummy ContextNotDefined Exception
        """
        pass

def exception_handler(exception_type, exception, traceback):
    """Exception handler. Writes stack trace to debug log in case an exception
    is not caught.
    """

    logging.error("An unexpected {} exception occurred: {}\nTraceback will be "
          "written to debug log (enable with --log-level debug)."
          .format(exception_type.__name__, _get_message(exception)))
    logging.debug("Traceback:")
    for item in format_tb(traceback):
        logging.debug(item.strip())
    sys.exit(2)

def main():
    """The main function of theanolm command.
    """

    parser = argparse.ArgumentParser(prog='theanolm')
    subparsers = parser.add_subparsers(
        title='commands',
        help='selects the command to perform ("theanolm command --help" '
             'displays help for the specific command)')

    train_parser = subparsers.add_parser(
        'train', help='train a model')
    train.add_arguments(train_parser)
    train_parser.set_defaults(command_function=train.train)

    score_parser = subparsers.add_parser(
        'score', help='score text or n-best lists using a model')
    score.add_arguments(score_parser)
    score_parser.set_defaults(command_function=score.score)

    decode_parser = subparsers.add_parser(
        'decode', help='decode a word lattice using a model')
    decode.add_arguments(decode_parser)
    decode_parser.set_defaults(command_function=decode.decode)

    sample_parser = subparsers.add_parser(
        'sample', help='generate text using a model')
    sample.add_arguments(sample_parser)
    sample_parser.set_defaults(command_function=sample.sample)

    version_parser = subparsers.add_parser(
        'version', help='display the version number')
    version_parser.set_defaults(command_function=version.version)

    args = parser.parse_args()

    if hasattr(args, 'command_function'):
        sys.excepthook = exception_handler
        try:
            # Increasing recursion limit is necessary with deep networks.
            sys.setrecursionlimit(100000)
            args.command_function(args)
        except FileNotFoundError as e:
            print('Could not open one of the required files. The error message '
                  'was:',
                  _get_message(e))
            sys.exit(2)
        except InputError as e:
            print('There was an error in one of the provided input files. The '
                  'error message was:',
                  _get_message(e))
            sys.exit(2)
        except NumberError as e:
            print('A numerical error has occurred. This may happen e.g. when '
                  'network parameters go to infinity. If this happens during '
                  'training, using a smaller learning rate usually helps. '
                  'Another possibility is to use gradient normalization. If '
                  'this happens during inference, there is probably something '
                  'wrong in the model. The error message was:',
                  _get_message(e))
            sys.exit(2)
        except ContextNotDefined as e:
            print('Theano returned error "{}". You need to map the device '
                  'contexts that are used in the architecture file to CUDA '
                  'devices '
                  '(e.g. THEANO_FLAGS=contexts=dev0->cuda0;dev1->cuda1).'
                  .format(_get_message(e)))
            sys.exit(2)
        except TheanoConfigurationError as e:
            print('There is a problem with Theano configuration. Please check '
                  'THEANO_FLAGS environment variable and .theanorc file. The '
                  'error message was:',
                  _get_message(e))
            sys.exit(2)
        except GpuArrayException as e:
            if 'out of memory' in _get_message(e):
                print('Theano could not fit all the required variables in the '
                      'memory. Memory consumption can be reduced by using a '
                      'smaller vocabulary (classes or shortlist), smaller '
                      'layers, smaller mini-batch size, shorter sequence '
                      'length, or dividing the model over multiple GPUs. '
                      'Traceback will be written to debug log.')
            else:
                print('Theano returned a gpuarray error "{}". Traceback will '
                      'be written to debug log.'
                      .format(_get_message(e)))
            logging.debug(e, exc_info=True)
            sys.exit(2)
        except IncompatibleStateError as e:
            print('The neural network state that you tried to load is invalid. '
                  'It may have been created with an old version of TheanoLM, '
                  'or the file may be corrupt. The error message was:',
                  _get_message(e))
            sys.exit(2)
        except KeyboardInterrupt:
            print('Interrupted from keyboard.')
    else:
        parser.print_help()

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the BasicOptimizer class, a base class for
optimizers.
"""

from abc import abstractmethod, ABCMeta

import theano
import theano.tensor as tensor

from theanolm.backend import IncompatibleStateError
from theanolm.backend import test_value
from theanolm.backend import l1_norm, sum_of_squares

class Cost(object, metaclass=ABCMeta):
    """Base Class for Cost Functions
    """

    def __init__(self, network, exclude_id=None, l1_regularization=None,
                 l2_regularization=None, epsilon=1e-6):
        """Constructs a cost function.

        :type network: Network
        :param network: the cost function will be defined with regard to the
                        outputs of this network

        :type exclude_id: int
        :param exclude_id: if other than ``None``, exclude these class IDs from
                           the cost (useful for excluding <unk> tokens)

        :type l1_regularization: float
        :param l1_regularization: if other than ``None``, the cost will include
            a term that is the L1 norm weighted by this value

        :type l2_regularization: float
        :param l2_regularization: if other than ``None``, the cost will include
            a term that is the L2 norm weighted by this value

        :type epsilon: float
        :param epsilon: numerical stability term, added to probabilities before
                        taking the logarithm
        """

        self._network = network
        self._exclude_id = exclude_id
        self._l1_regularization = l1_regularization
        self._l2_regularization = l2_regularization
        self._epsilon = epsilon

    @abstractmethod
    def _get_logprobs(self):
        """Returns a symbolic variable that represents the log probabilities of
        the target outputs, computed either exactly or using some approximation.

        :rtype: symbolic 2D tensor
        :returns: the log probability of each time step of each sequence
        """

        assert False

    def get_tensor(self):
        """Returns a symbolic variable that represents the mini-batch cost.

        The cost is the negative log probability normalized by the number of
        tokens in the mini-batch, so that the gradients will also be normalized
        by the number of tokens. Tokens that are outside the mask will be
        excluded from the cost. If ``exclude_id`` is given to the constructor,
        that class ID will be excluded too.

        If ``l1_regularization`` or ``l2_regularization`` is given to the
        constructor, a term will be added to the cost that penalizes for large
        norm accordingly.

        :rtype: a tuple of a symbolic 2D tensor and a symbolic integer
        :returns: the cost value for each time step of each sequence, and the
                  number of words in the mini-batch
        """

        mask = self._network.mask
        if self._exclude_id is not None:
            mask *= tensor.neq(self._network.target_word_ids, self._exclude_id)
        # The mask has to be cast to floatX, otherwise the result will be
        # float64 and pulled out from the GPU earlier than necessary.
        logprobs = self._get_logprobs() * tensor.cast(mask, theano.config.floatX)
        num_words = mask.sum()
        cost = -logprobs.sum() / num_words

        model_params = self._network.get_variables().values()
        l1_weight = self._l1_regularization
        l2_weight = self._l2_regularization
        if l1_weight is not None:
            cost += l1_weight * l1_norm(model_params)
        if l2_weight is not None:
            cost += l2_weight * sum_of_squares(model_params)

        return cost, num_words

class CrossEntropyCost(Cost):
    """Cross-Entropy Cost
    """

    def __init__(self, *args, **kwargs):
        """Constructs a cost function.

        :type network: Network
        :param network: the cost function will be defined with regard to the
                        outputs of this network

        :type exclude_id: int
        :param exclude_id: if other than ``None``, exclude these class IDs from
                           the cost (useful for excluding <unk> tokens)

        :type epsilon: float
        :param epsilon: numerical stability term, added to probabilities before
                        taking the logarithm
        """

        super().__init__(*args, **kwargs)

    def _get_logprobs(self):
        """Returns a symbolic variable that represents the log probabilities of
        the target outputs.

        :rtype: Variable
        :returns: a symbolic 2-dimensional matrix that contains the log
                  probability of each time step of each sequence
        """

        return tensor.log(self._network.target_probs())

class NCECost(Cost):
    """Noise-Contrastive Estimation Cost

    M. U. Gutmann (2012)
    Noise-Contrastive Estimation of Unnormalized Statistical Models, with
    Applications to Natural Image Statistics
    http://www.jmlr.org/papers/v13/gutmann12a.html
    """

    def __init__(self, *args, **kwargs):
        """Constructs a cost function.

        :type network: Network
        :param network: the cost function will be defined with regard to the
                        outputs of this network

        :type exclude_id: int
        :param exclude_id: if other than ``None``, exclude these class IDs from
                           the cost (useful for excluding <unk> tokens)

        :type epsilon: float
        :param epsilon: numerical stability term, added to probabilities before
                        taking the logarithm
        """

        super().__init__(*args, **kwargs)

    def _get_logprobs(self):
        """Returns a symbolic variable that represents the log probabilities of
        the target outputs.

        :rtype: Variable
        :returns: a symbolic 2-dimensional matrix that contains the log
                  probability of each time step of each sequence
        """

        target_logprobs = self._network.unnormalized_logprobs()
        target_class_ids = self._network.target_class_ids
        noise_distribution = self._network.noise_distribution
        # If a single value is retured, it will be broadcasted to the mini-batch
        # shape.
        target_prior_probs = noise_distribution.probs(target_class_ids)
        target_prior_logprobs = tensor.log(target_prior_probs + self._epsilon)
        # In the article, h = 1 / (1 + e^-G). log(h) can be expressed using the
        # softplus function: log(h) = -log(1 + e^-G) = -softplus(-G)
        G = target_logprobs - target_prior_logprobs
        target_log_h = -tensor.nnet.softplus(-G)

        sample, sample_logprobs = self._network.noise_sample()
        # sample_prior_logprobs will be a one-dimensional array (or a scalar in
        # case of uniform noise), but it will be broadcasted when subtracted
        # from sample_logprobs.
        sample_prior_probs = noise_distribution.probs(sample)
        sample_prior_logprobs = tensor.log(sample_prior_probs + self._epsilon)
        # log(1 - h) = log(1 - e^G / (e^G + 1))
        #            = log((e^G + 1 - e^G) / (e^G + 1))
        #            = log(1) - log(e^G + 1)
        #            = -softplus(G)
        G = sample_logprobs - sample_prior_logprobs
        sample_log_one_minus_h = -tensor.nnet.softplus(G)
        return target_log_h + sample_log_one_minus_h.sum(2)

class BlackoutCost(Cost):
    """BlackOut Cost

    S. Ji (2016)
    BlackOut: Speeding up Recurrent Neural Network Language Models With Very
    Large Vocabularies
    https://arxiv.org/abs/1511.06909
    """

    def __init__(self, *args, **kwargs):
        """Constructs a cost function.

        :type network: Network
        :param network: the cost function will be defined with regard to the
                        outputs of this network

        :type exclude_id: int
        :param exclude_id: if other than ``None``, exclude these class IDs from
                           the cost (useful for excluding <unk> tokens)

        :type epsilon: float
        :param epsilon: numerical stability term, added to probabilities before
                        taking the logarithm
        """

        super().__init__(*args, **kwargs)

    def _get_logprobs(self):
        """Returns a symbolic variable that represents the log probabilities of
        the target outputs.

        :rtype: Variable
        :returns: a symbolic 2-dimensional matrix that contains the log
                  probability of each time step of each sequence
        """

        target_logprobs = self._network.unnormalized_logprobs()
        target_probs = tensor.exp(target_logprobs)
        target_class_ids = self._network.target_class_ids
        noise_distribution = self._network.noise_distribution
        # If a single value is retured, it will be broadcasted to the mini-batch
        # shape.
        target_prior_probs = noise_distribution.probs(target_class_ids)
        target_weighted_probs = target_probs / target_prior_probs

        sample, sample_logprobs = self._network.noise_sample()
        # sample_prior_probs will be a one-dimensional array (or a scalar in
        # case of uniform noise), but it will be broadcasted when used to divide
        # sample_logprobs.
        sample_probs = tensor.exp(sample_logprobs)
        sample_prior_probs = noise_distribution.probs(sample)
        sample_weighted_probs = sample_probs / sample_prior_probs

        denominators = target_weighted_probs + \
                       sample_weighted_probs.sum(2)
        target_costs = target_weighted_probs / denominators
        sample_costs = sample_weighted_probs / denominators[:, :, None]
        sample_costs = 1.0 - sample_costs
        result = tensor.log(target_costs + self._epsilon)
        result += tensor.log(sample_costs + self._epsilon).sum(2)
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the dropout layer.
"""

import logging

import numpy
import theano
import theano.tensor as tensor

from theanolm.network.basiclayer import BasicLayer

class DropoutLayer(BasicLayer):
    """Dropout Layer

    A dropout layer is not a regular layer in the sense that it doesn't contain
    any neurons. It simply randomly sets some activations to zero at train time
    to prevent overfitting.

    N. Srivastava et al. (2014)
    Dropout: A Simple Way to Prevent Neural Networks from Overfitting
    Journal of Machine Learning Research, 15, 1929-1958
    """

    def __init__(self, layer_options, *args, **kwargs):
        """Validates the parameters of this layer.
        """

        super().__init__(layer_options, *args, **kwargs)

        if 'dropout_rate' in layer_options:
            self._dropout_rate = float(layer_options['dropout_rate'])
            # Don't allow dropout rate too close to 1.0 to prevent division by
            # zero.
            self._dropout_rate = min(0.9999, self._dropout_rate)
        else:
            self._dropout_rate = 0.5
        logging.debug("  dropout_rate=%f", self._dropout_rate)

        # Make sure the user hasn't tried to change the number of connections.
        input_size = sum(x.output_size for x in self._input_layers)
        output_size = self.output_size
        if input_size != output_size:
            raise ValueError("Dropout layer size has to match the previous "
                             "layer.")

        self.output = None

    def create_structure(self):
        """Creates the symbolic graph of this layer.

        Sets ``self.output`` to a symbolic matrix that describes the output of
        this layer. During training sets randomly some of the outputs to zero.
        """

        float_type = numpy.dtype(theano.config.floatX).type

        layer_input = tensor.concatenate([x.output for x in self._input_layers],
                                         axis=2)
        # Pass rate is the probability of not dropping a unit.
        pass_rate = 1.0 - self._dropout_rate
        pass_rate = float_type(pass_rate)
        sample = self._network.random.uniform(size=layer_input.shape)
        mask = tensor.cast(sample < pass_rate, theano.config.floatX)
        # Multiply the output by the inverse of the pass rate before dropping
        # units to compensate the scaling effect.
        scale_correction = 1.0 / pass_rate
        scale_correction = float_type(scale_correction)
        self.output = tensor.switch(self._network.is_training,
                                    layer_input * scale_correction * mask,
                                    layer_input)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import argparse
import logging
from time import time
from theanolm.backend import TextFileType
from wordclasses import TheanoBigramOptimizer, NumpyBigramOptimizer
from theanolm.vocabulary import Vocabulary
from theanolm.vocabulary import compute_word_counts, BigramStatistics
from wordclasses.functions import is_scheduled

def save(optimizer, output_file, output_format):
    """Writes the current classes to a file.

    If the output file is seekable, first rewinds and truncates the file.

    :type optimizer: BigramOptimizer
    :param optimizer: save the current state of this optimizer

    :type output_file: file object
    :param output_file: a file or stream where to save the classes

    :type output_format: str
    :param output_format: either "classes" or "srilm-classes" - selects the
                          output file format
    """

    if output_file.seekable():
        output_file.seek(0)
        output_file.truncate()

    for word, class_id, prob in optimizer.words():
        if output_format == 'classes':
            output_file.write('{} {}\n'.format(word, class_id))
        elif output_format == 'srilm-classes':
            output_file.write('CLASS-{:05d} {} {}\n'.format(class_id, prob, word))

def main():
    parser = argparse.ArgumentParser(prog='wctool')

    argument_group = parser.add_argument_group("files")
    argument_group.add_argument(
        '--training-set', metavar='FILE', type=TextFileType('r'),
        nargs='+', required=True,
        help='text or .gz files containing training data (one sentence per '
             'line)')
    argument_group.add_argument(
        '--vocabulary', metavar='FILE', type=TextFileType('r'), default=None,
        help='text or .gz file containing a list of words to include in class '
             'forming, and possibly their initial classes')
    argument_group.add_argument(
        '--vocabulary-format', metavar='FORMAT', type=str, default='words',
        help='vocabulary format, one of "words" (one word per line, default), '
             '"classes" (word and class ID per line), "srilm-classes" (class '
             'name, membership probability, and word per line)')
    argument_group.add_argument(
        '--output-file', metavar='FILE', type=TextFileType('w'), default='-',
        help='where to write the word classes (default stdout)')
    argument_group.add_argument(
        '--output-format', metavar='FORMAT', type=str, default='srilm-classes',
        help='format of the output file, one of "classes" (word and class ID '
             'per line), "srilm-classes" (default; class name, membership '
             'probability, and word per line)')
    argument_group.add_argument(
        '--output-frequency', metavar='N', type=int, default='1',
        help='save classes N times per optimization iteration (default 1)')

    argument_group = parser.add_argument_group("optimization")
    argument_group.add_argument(
        '--num-classes', metavar='N', type=int, default=2000,
        help='number of classes to form, if vocabulary is not specified '
             '(default 2000)')
    argument_group.add_argument(
        '--method', metavar='NAME', type=str, default='bigram-theano',
        help='method for creating word classes, one of "bigram-theano", '
             '"bigram-numpy" (default "bigram-theano")')

    argument_group = parser.add_argument_group("logging and debugging")
    argument_group.add_argument(
        '--log-file', metavar='FILE', type=str, default='-',
        help='path where to write log file (default is standard output)')
    argument_group.add_argument(
        '--log-level', metavar='LEVEL', type=str, default='info',
        help='minimum level of events to log, one of "debug", "info", "warn" '
             '(default "info")')
    argument_group.add_argument(
        '--log-interval', metavar='N', type=int, default=1000,
        help='print statistics after every Nth word; quiet if less than one '
             '(default 1000)')

    args = parser.parse_args()

    log_file = args.log_file
    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        raise ValueError("Invalid logging level requested: " + args.log_level)
    log_format = '%(asctime)s %(funcName)s: %(message)s'
    if args.log_file == '-':
        logging.basicConfig(stream=sys.stdout, format=log_format, level=log_level)
    else:
        logging.basicConfig(filename=log_file, format=log_format, level=log_level)

    if args.vocabulary is None:
        word_counts = compute_word_counts(args.training_set)
        vocabulary = Vocabulary.from_word_counts(word_counts,
                                                 args.num_classes)
        for subset_file in args.training_set:
            subset_file.seek(0)
    else:
        vocabulary = Vocabulary.from_file(args.vocabulary,
                                          args.vocabulary_format)

    print("Number of words in vocabulary:", vocabulary.num_shortlist_words())
    print("Number of word classes:", vocabulary.num_classes())
    print("Number of normal word classes:", vocabulary.num_normal_classes)

    logging.info("Reading word unigram and bigram statistics.")
    statistics = BigramStatistics(args.training_set, vocabulary)

    if args.method == 'bigram-theano':
        optimizer = TheanoBigramOptimizer(statistics, vocabulary)
    elif args.method == 'bigram-numpy':
        optimizer = NumpyBigramOptimizer(statistics, vocabulary)
    else:
        raise ValueError("Invalid method requested: " + args.method)

    iteration = 1
    while True:
        logging.info("Starting iteration %d.", iteration)
        num_words = 0
        num_moves = 0
        for word in vocabulary.words():
            start_time = time()
            num_words += 1
            if optimizer.move_to_best_class(word):
                num_moves += 1
            duration = time() - start_time
            if (args.log_interval >= 1) and \
               (num_words % args.log_interval == 0):
                logging.info("[%d] (%.1f %%) of iteration %d -- moves = %d, cost = %.2f, duration = %.1f ms",
                     num_words,
                     num_words / vocabulary.num_shortlist_words() * 100,
                     iteration,
                     num_moves,
                     optimizer.log_likelihood(),
                     duration * 100)
            if is_scheduled(num_words,
                            args.output_frequency,
                            vocabulary.num_shortlist_words()):
                save(optimizer, args.output_file, args.output_format)

        if num_moves == 0:
            break
        iteration += 1

    logging.info("Optimization finished.")
    save(optimizer, args.output_file, args.output_format)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Implements the softmax output layer.
"""

import numpy
import theano
from theano import tensor

from theanolm.network.samplingoutputlayer import SamplingOutputLayer

class SoftmaxLayer(SamplingOutputLayer):
    """Softmax Output Layer

    The output layer is a simple softmax layer that outputs the word
    probabilities.
    """

    def __init__(self, *args, **kwargs):
        """Initializes the parameters used by this layer.
        """

        super().__init__(*args, **kwargs)

        # Create the parameters. Weight matrix and bias for each input.
        input_size = sum(x.output_size for x in self._input_layers)
        output_size = self.output_size
        self._init_weight('input/W', (input_size, output_size), scale=0.01)
        if self._network.class_prior_probs is None:
            self._init_bias('input/b', output_size)
        else:
            initial_bias = numpy.log(self._network.class_prior_probs + 1e-10)
            self._init_bias('input/b', output_size, initial_bias)

        vocabulary = self._network.vocabulary
        self._unk_id = vocabulary.word_to_id['<unk>']
        self._unk_class_id = vocabulary.word_id_to_class_id[self._unk_id]

        self.output_probs = None
        self.target_probs = None
        self.unnormalized_logprobs = None
        self._layer_input = None

    def create_structure(self):
        """Creates the symbolic graph of this layer.

        The input is always 3-dimensional: the first dimension is the time step,
        the second dimension are the sequences, and the third dimension is the
        word projection. When generating text, there's just one sequence and one
        time step in the input.

        Sets ``self.output_probs`` to a symbolic matrix that specifies output
        probabilities for all classes, ``self.target_probs`` to one that
        specifies output probabilities for the target classes, and
        ``self.unnormalized_logprobs``, ``self.sample``,
        ``self.sample_logprobs``, ``self.seqshared_sample``,
        ``self.seqshared_sample_logprobs``, ``self.shared_sample``, and
        ``self.shared_sample_logprobs`` to symbolic matrices that can be used to
        obtain the data required to compute sampling output costs.
        """

        layer_input = tensor.concatenate([x.output for x in self._input_layers],
                                         axis=2)
        preact = self._tensor_preact(layer_input, 'input')

        # Combine the first two dimensions so that softmax is taken
        # independently for each location, over the output classes. If <unk> is
        # excluded, set those activations to -inf before normalization.
        # NOTE: We're checking the class ID so this will fail if there are other
        # words in the same class with <unk> word!
        num_time_steps = layer_input.shape[0]
        num_sequences = layer_input.shape[1]
        preact = preact.reshape([num_time_steps * num_sequences,
                                 self.output_size])
        if self._network.exclude_unk:
            float_type = numpy.dtype(theano.config.floatX).type
            log_zero = float_type('-inf')
            preact = tensor.set_subtensor(preact[:, self._unk_class_id],
                                          log_zero)
        output_probs = tensor.nnet.softmax(preact)

        # This variable contains probabilities for the whole vocabulary.
        self.output_probs = output_probs.reshape([num_time_steps,
                                                  num_sequences,
                                                  self.output_size])

        # The following variables can only be used when
        # self._network.target_class_ids is given to the function.
        element_ids = tensor.arange(num_time_steps * num_sequences)
        target_class_ids = self._network.target_class_ids.flatten()
        target_probs = output_probs[(element_ids, target_class_ids)]
        self.target_probs = target_probs.reshape([num_time_steps,
                                                  num_sequences])

        # Define layer input and unnormalized logprobs for computing
        # sampling-based costs. Only possible if the parameters are not split to
        # multiple devices.
        self._layer_input = layer_input
        if not len(self._devices) > 1:
            self.unnormalized_logprobs = \
                self._get_unnormalized_logprobs()



from theanolm.backend import TextFileType
from theanolm.vocabulary.statistics import BigramStatistics
from wordclasses.theanobigramoptimizer import TheanoBigramOptimizer
from wordclasses.numpybigramoptimizer import NumpyBigramOptimizer
from wordclasses.wctool import main



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import numpy
from wordclasses.bigramoptimizer import BigramOptimizer
from wordclasses.functions import byte_size

class NumpyBigramOptimizer(BigramOptimizer):
    """Word Class Optimizer
    """

    def __init__(self, statistics, vocabulary):
        """Computes initial statistics.

        :type statistics: BigramStatistics
        :param statistics: word statistics from the training corpus

        :type vocabulary: theanolm.Vocabulary
        :param vocabulary: words to include in the optimization and initial classes
        """

        # count_nonzero() and any() seem to fail on the sparse matrix.
        if not statistics.unigram_counts.any():
            raise ValueError("Empty word unigram statistics.")
        if statistics.bigram_counts.nnz == 0:
            raise ValueError("Empty word bigram statistics.")

        super().__init__(vocabulary)

        # Create word counts.
        self._word_counts = statistics.unigram_counts
        logging.debug("Allocated %s for word counts.",
                      byte_size(self._word_counts.nbytes))
        self._ww_counts = statistics.bigram_counts.tocsc()
        logging.debug("Allocated %s for sparse word-word counts.",
                      byte_size(self._ww_counts.data.nbytes))

        # Initialize classes.
        self._word_to_class = numpy.array(vocabulary.word_id_to_class_id)
        logging.debug("Allocated %s for word-to-class mapping.",
                      byte_size(self._word_to_class.nbytes))

        # Compute class counts from word counts.
        logging.info("Computing class and class/word statistics.")
        self._class_counts, self._cc_counts, self._cw_counts, self._wc_counts = \
            self._compute_class_statistics(self._word_counts,
                                           self._ww_counts,
                                           self._word_to_class)

    def get_word_class(self, word_id):
        """Returns the class the given word is currently assigned to.

        :type word_id: int
        :param word_id: ID of the word

        :rtype: int
        :returns: ID of the class the word is assigned to
        """

        return self._word_to_class[word_id]

    def get_word_prob(self, word_id):
        """Returns the unigram probability of a word within its class.

        :type word_id: int
        :param word_id: ID of the word

        :rtype: float
        :returns: class membership probability
        """

        word_count = self._word_counts[word_id]
        class_id = self._word_to_class[word_id]
        class_count = self._class_counts[class_id]
        if class_count == 0:
            return 0.0
        else:
            return word_count / class_count

    def log_likelihood(self):
        """Computes the log likelihood that a bigram model would give to the
        corpus.

        :rtype: float
        :returns: log likelihood of the training corpus
        """

        return (numpy.ma.log(self._cc_counts) * self._cc_counts).sum() + \
               (numpy.ma.log(self._word_counts) * self._word_counts).sum() - \
               2 * (numpy.ma.log(self._class_counts) * self._class_counts).sum()

    def _evaluate(self, word_id, new_class_id):
        """Evaluates how much moving a word to another class would change the
        log likelihood.
        """

        old_class_id = self.get_word_class(word_id)

        # old class
        old_count = self._class_counts[old_class_id]
        new_count = old_count - self._word_counts[word_id]
        result = 2 * old_count * numpy.log(old_count)
        result -= 2 * new_count * numpy.log(new_count)

        # new class
        old_count = self._class_counts[new_class_id]
        new_count = old_count + self._word_counts[word_id]
        result += 2 * old_count * numpy.log(old_count)
        result -= 2 * new_count * numpy.log(new_count)

        # Iterate over classes other than the old and new class of the word.
        class_ids = numpy.arange(self.num_classes)
        selector = (class_ids != old_class_id) & (class_ids != new_class_id)
        iter_class_ids = class_ids[selector]

        # old class, class X
        old_counts = self._cc_counts[old_class_id,iter_class_ids]
        new_counts = old_counts - self._wc_counts[word_id,iter_class_ids]
        result -= (numpy.ma.log(old_counts) * old_counts).sum()
        result += (numpy.ma.log(new_counts) * new_counts).sum()

        # new class, class X
        old_counts = self._cc_counts[new_class_id,iter_class_ids]
        new_counts = old_counts + self._wc_counts[word_id,iter_class_ids]
        result -= (numpy.ma.log(old_counts) * old_counts).sum()
        result += (numpy.ma.log(new_counts) * new_counts).sum()

        # class X, old class
        old_counts = self._cc_counts[iter_class_ids,old_class_id]
        new_counts = old_counts - self._cw_counts[iter_class_ids,word_id]
        result -= (numpy.ma.log(old_counts) * old_counts).sum()
        result += (numpy.ma.log(new_counts) * new_counts).sum()

        # class X, new class
        old_counts = self._cc_counts[iter_class_ids,new_class_id]
        new_counts = old_counts + self._cw_counts[iter_class_ids,word_id]
        result -= (numpy.ma.log(old_counts) * old_counts).sum()
        result += (numpy.ma.log(new_counts) * new_counts).sum()

        # old class, new class
        old_count = self._cc_counts[old_class_id,new_class_id]
        new_count = old_count - \
                    self._wc_counts[word_id,new_class_id] + \
                    self._cw_counts[old_class_id,word_id] - \
                    self._ww_counts[word_id,word_id]
        result += self._ll_change(old_count, new_count)

        # new class, old class
        old_count = self._cc_counts[new_class_id,old_class_id]
        new_count = old_count - \
                    self._cw_counts[new_class_id,word_id] + \
                    self._wc_counts[word_id,old_class_id] - \
                    self._ww_counts[word_id,word_id]
        result += self._ll_change(old_count, new_count)

        # old class, old class
        old_count = self._cc_counts[old_class_id,old_class_id]
        new_count = old_count - \
                    self._wc_counts[word_id,old_class_id] - \
                    self._cw_counts[old_class_id,word_id] + \
                    self._ww_counts[word_id,word_id]
        result += self._ll_change(old_count, new_count)

        # new class, new class
        old_count = self._cc_counts[new_class_id,new_class_id]
        new_count = old_count + \
                    self._wc_counts[word_id,new_class_id] + \
                    self._cw_counts[new_class_id,word_id] + \
                    self._ww_counts[word_id,word_id]
        result += self._ll_change(old_count, new_count)

        return result

    def _ll_change(self, old_count, new_count):
        result = 0
        if old_count != 0:
            result -= old_count * numpy.log(old_count)
        if new_count != 0:
            result += new_count * numpy.log(new_count)
        return result

    def _move(self, word_id, new_class_id):
        """Moves a word to another class.
        """

        old_class_id = self._word_to_class[word_id]

        # word
        word_count = self._word_counts[word_id]
        self._class_counts[old_class_id] -= word_count
        self._class_counts[new_class_id] += word_count

        # word, word X
        right_word_ids = numpy.asarray(
            [id for id in self._ww_counts[word_id,:].nonzero()[1] if id != word_id])
        right_class_ids = self._word_to_class[right_word_ids]
        counts = self._ww_counts[word_id,right_word_ids].toarray().flatten()
        self._cw_counts[old_class_id,right_word_ids] -= counts
        self._cw_counts[new_class_id,right_word_ids] += counts
        numpy.add.at(self._cc_counts[old_class_id,:], right_class_ids, -counts)
        numpy.add.at(self._cc_counts[new_class_id,:], right_class_ids, counts)

        # word X, word
        left_word_ids = numpy.asarray(
            [id for id in self._ww_counts[:,word_id].nonzero()[0] if id != word_id])
        left_class_ids = self._word_to_class[left_word_ids]
        counts = self._ww_counts[left_word_ids,word_id].toarray().flatten()
        self._wc_counts[left_word_ids,old_class_id] -= counts
        self._wc_counts[left_word_ids,new_class_id] += counts
        numpy.add.at(self._cc_counts[:,old_class_id], left_class_ids, -counts)
        numpy.add.at(self._cc_counts[:,new_class_id], left_class_ids, counts)

        # word, word
        count = self._ww_counts[word_id,word_id]
        self._cc_counts[old_class_id,old_class_id] -= count
        self._cc_counts[new_class_id,new_class_id] += count
        self._cw_counts[old_class_id,word_id] -= count
        self._cw_counts[new_class_id,word_id] += count
        self._wc_counts[word_id,old_class_id] -= count
        self._wc_counts[word_id,new_class_id] += count

        self._word_to_class[word_id] = new_class_id

    def _class_size(self, class_id):
        """Calculates the number of words in a class.

        :type class_id: int
        :param class_id: ID of a class

        :rtype: int
        :returns: number of words in the class
        """

        return numpy.count_nonzero(self._word_to_class == class_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Functions for creating matrices and accessing submatrices.
"""

import numpy
import theano

def random_normal_matrix(shape, scale=None):
    """Generates a random matrix from “standard normal” distribution.

    :type shape: tuple of ints
    :param shape: size of each dimension (typically there are two dimensions,
                  input and output)

    :type scale: float
    :param scale: if other than None, the random numbers will be scaled by this
                  factor

    :rtype: numpy.ndarray
    :returns: the generated weight matrix
    """

    result = numpy.random.randn(*shape)
    if scale is not None:
        result = scale * result
    return result.astype(theano.config.floatX)

def random_orthogonal_matrix(size):
    """Generates a random orthogonal matrix from “standard normal” distribution.

    :type size: int
    :param size: size of both dimensions of the weight

    :rtype: numpy.ndarray
    :returns: the generated weight matrix
    """

    nonorthogonal_matrix = numpy.random.randn(size, size)
    result, _, _ = numpy.linalg.svd(nonorthogonal_matrix)
    return result.astype(theano.config.floatX)

def random_matrix(shape, scale=None, count=1):
    """Generates a random matrix from “standard normal” distribution.

    If ``shape`` contains two dimensions that match, generates an orthogonal
    matrix. In that case scale is ignored.

    If ``count`` is specified, creates a concatenation of several similar
    matrices (same shape but different content).

    :type shape: tuple of ints
    :param shape: size of each dimension (typically there are two dimensions,
                  input and output)

    :type scale: float
    :param scale: if other than None, the random numbers will be scaled by this
                  factor

    :type count: int
    :param count: concatenate this many weight matrices with the same shape

    :rtype: numpy.ndarray
    :returns: the generated weight matrix
    """

    if (len(shape) == 2) and (shape[0] == shape[1]):
        return numpy.concatenate(
            [random_orthogonal_matrix(shape[0]) for _ in range(count)],
            axis=-1)
    else:
        return numpy.concatenate(
            [random_normal_matrix(shape, scale) for _ in range(count)],
            axis=-1)

def matrix_from_value(shape, value=None):
    """Creates a matrix with given value.

    If ``value`` is not given, initializes the vector with zero value. If
    ``value`` is a list, creates a concatenation of as many vectors as there are
    elements in the list.

    :type shape: int or tuple of ints
    :param shape: size of the vector, or a tuple of the sizes of each dimension
                  (in case ``value`` is a list, each part will have this size)

    :type value: float, numpy.ndarray or list
    :param value: the value or array to initialize the elements to, or a list of
                  values or arrays to create a concatenation of vectors
    """

    values = value if isinstance(value, (list, tuple)) else [value]
    parts = []
    for part_value in values:
        if part_value is None:
            part = numpy.zeros(shape).astype(theano.config.floatX)
        elif isinstance(value, numpy.ndarray):
            part = value.astype(theano.config.floatX)
        else:
            part = numpy.empty(shape).astype(theano.config.floatX)
            part.fill(part_value)
        parts.append(part)
    return numpy.concatenate(parts, axis=-1)

def get_submatrix(matrices, index, size, end_index=None):
    """Returns a submatrix of a concatenation of 2 or 3 dimensional
    matrices.

    :type matrices: Variable
    :param matrices: symbolic 2 or 3 dimensional matrix formed by
                     concatenating matrices of length size

    :type index: int
    :param index: index of the matrix to be returned

    :type size: Variable
    :param size: size of the last dimension of one submatrix

    :type end_index: int
    :param end_index: if set to other than None, returns a concatenation of all
                      the submatrices from ``index`` to ``end_index``
    """

    if end_index is None:
        end_index = index
    start = index * size
    end = (end_index + 1) * size
    if matrices.ndim == 3:
        return matrices[:, :, start:end]
    elif matrices.ndim == 2:
        return matrices[:, start:end]
    else:
        raise ValueError("get_submatrix() requires a 2 or 3 dimensional matrix.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A package that provides functions and classes related to basic computation.
"""

from theanolm.backend.exceptions import *
from theanolm.backend.filetypes import TextFileType
from theanolm.backend.gpu import get_default_device, log_free_mem
from theanolm.backend.parameters import Parameters
from theanolm.backend.classdistribution import UniformDistribution
from theanolm.backend.classdistribution import LogUniformDistribution
from theanolm.backend.classdistribution import MultinomialDistribution
from theanolm.backend.matrixfunctions import test_value
from theanolm.backend.probfunctions import interpolate_linear
from theanolm.backend.probfunctions import interpolate_loglinear
from theanolm.backend.probfunctions import logprob_type
from theanolm.backend.operations import conv1d, conv2d
from theanolm.backend.operations import l1_norm, sum_of_squares



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the "theanolm score" command.
"""

import sys
import logging

import numpy
import theano

from theanolm import Network
from theanolm.backend import TextFileType, get_default_device
from theanolm.parsing import ScoringBatchIterator
from theanolm.scoring import TextScorer

def add_arguments(parser):
    """Specifies the command line arguments supported by the "theanolm score"
    command.

    :type parser: argparse.ArgumentParser
    :param parser: a command line argument parser
    """

    argument_group = parser.add_argument_group("files")
    argument_group.add_argument(
        'model_path', metavar='MODEL-FILE', type=str,
        help='the model file that will be used to score text')
    argument_group.add_argument(
        'input_file', metavar='TEXT-FILE', type=TextFileType('r'),
        help='text file containing text to be scored (UTF-8, one sentence per '
             'line, assumed to be compressed if the name ends in ".gz")')
    argument_group.add_argument(
        '--output-file', metavar='FILE', type=TextFileType('w'), default='-',
        help='where to write the statistics (default stdout, will be '
             'compressed if the name ends in ".gz")')

    argument_group = parser.add_argument_group("scoring")
    argument_group.add_argument(
        '--output', metavar='DETAIL', type=str, default='perplexity',
        choices=['perplexity', 'utterance-scores', 'word-scores'],
        help='what to output, one of "perplexity", "utterance-scores", '
             '"word-scores" (default "perplexity")')
    argument_group.add_argument(
        '--log-base', metavar='B', type=int, default=None,
        help='convert output log probabilities to base B (default is the '
             'natural logarithm)')
    argument_group.add_argument(
        '--exclude-unk', action="store_true",
        help="exclude <unk> tokens from perplexity computation")
    argument_group.add_argument(
        '--subwords', metavar='MARKING', type=str, default=None,
        choices=['word-boundary', 'prefix-affix', None],
        help='the subword vocabulary uses MARKING to indicate how words are '
             'formed from subwords; one of "word-boundary" (<w> token '
             'separates words), "prefix-affix" (subwords that can be '
             'concatenated are prefixed or affixed with +, e.g. "cat+ +s")')
    argument_group.add_argument(
        '--shortlist', action="store_true",
        help='distribute <unk> token probability among the out-of-shortlist '
             'words according to their unigram frequencies in the training '
             'data')

    argument_group = parser.add_argument_group("configuration")
    argument_group.add_argument(
        '--default-device', metavar='DEVICE', type=str, default=None,
        help='when multiple GPUs are present, use DEVICE as default')

    argument_group = parser.add_argument_group("logging and debugging")
    argument_group.add_argument(
        '--log-file', metavar='FILE', type=str, default='-',
        help='path where to write log file (default is standard output)')
    argument_group.add_argument(
        '--log-level', metavar='LEVEL', type=str, default='info',
        choices=['debug', 'info', 'warn'],
        help='minimum level of events to log, one of "debug", "info", "warn" '
             '(default "info")')
    argument_group.add_argument(
        '--debug', action="store_true",
        help='use test values to get better error messages from Theano')
    argument_group.add_argument(
        '--profile', action="store_true",
        help='enable profiling Theano functions')

def score(args):
    """A function that performs the "theanolm score" command.

    :type args: argparse.Namespace
    :param args: a collection of command line arguments
    """

    log_file = args.log_file
    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        print("Invalid logging level requested:", args.log_level)
        sys.exit(1)
    log_format = '%(asctime)s %(funcName)s: %(message)s'
    if args.log_file == '-':
        logging.basicConfig(stream=sys.stdout, format=log_format, level=log_level)
    else:
        logging.basicConfig(filename=log_file, format=log_format, level=log_level)

    if args.debug:
        theano.config.compute_test_value = 'warn'
        logging.info("Enabled computing test values for tensor variables.")
        logging.warning("GpuArray backend will fail random number generation!")
    else:
        theano.config.compute_test_value = 'off'
    theano.config.profile = args.profile
    theano.config.profile_memory = args.profile

    default_device = get_default_device(args.default_device)
    network = Network.from_file(args.model_path, exclude_unk=args.exclude_unk,
                                default_device=default_device)

    logging.info("Building text scorer.")
    scorer = TextScorer(network, args.shortlist, args.exclude_unk, args.profile)

    logging.info("Scoring text.")
    if args.output == 'perplexity':
        _score_text(args.input_file, network.vocabulary, scorer,
                    args.output_file, args.log_base, args.subwords, False)
    elif args.output == 'word-scores':
        _score_text(args.input_file, network.vocabulary, scorer,
                    args.output_file, args.log_base, args.subwords, True)
    elif args.output == 'utterance-scores':
        _score_utterances(args.input_file, network.vocabulary, scorer,
                          args.output_file, args.log_base)
    else:
        print("Invalid output format requested:", args.output)
        sys.exit(1)

def _score_text(input_file, vocabulary, scorer, output_file,
                log_base=None, subword_marking=None, word_level=False):
    """Reads text from ``input_file``, computes perplexity using
    ``scorer``, and writes to ``output_file``.

    :type input_file: file object
    :param input_file: a file that contains the input sentences in SRILM n-best
                       format

    :type vocabulary: Vocabulary
    :param vocabulary: vocabulary that provides mapping between words and word
                       IDs

    :type scorer: TextScorer
    :param scorer: a text scorer for rescoring the input sentences

    :type output_file: file object
    :param output_file: a file where to write the output n-best list in SRILM
                        format

    :type log_base: int
    :param log_base: if set to other than None, convert log probabilities to
                     this base

    :type subword_marking: str
    :param subword_marking: if other than None, vocabulary is subwords;
        "word-boundary" indicates <w> token separates words, "prefix-affix"
        indicates subwords are prefixed/affixed with +

    :type word_level: bool
    :param word_level: if set to True, also writes word-level statistics
    """

    scoring_iter = \
        ScoringBatchIterator(input_file,
                             vocabulary,
                             batch_size=16,
                             max_sequence_length=None,
                             map_oos_to_unk=False)
    log_scale = 1.0 if log_base is None else numpy.log(log_base)

    total_logprob = 0.0
    num_sentences = 0
    num_tokens = 0
    num_words = 0
    num_probs = 0
    num_unks = 0
    num_zeroprobs = 0
    for word_ids, words, mask in scoring_iter:
        class_ids, membership_probs = vocabulary.get_class_memberships(word_ids)
        logprobs = scorer.score_batch(word_ids, class_ids, membership_probs,
                                      mask)
        for seq_index, seq_logprobs in enumerate(logprobs):
            seq_word_ids = word_ids[:, seq_index]
            seq_mask = mask[:, seq_index]
            seq_word_ids = seq_word_ids[seq_mask == 1]
            seq_words = words[seq_index]
            merged_words, merged_logprobs = _merge_subwords(seq_words,
                                                            seq_logprobs,
                                                            subword_marking)

            # total logprob of this sequence
            seq_logprob = sum(lp for lp in merged_logprobs
                              if (lp is not None) and (not numpy.isneginf(lp)))
            # total logprob of all sequences
            total_logprob += seq_logprob
            # number of tokens, which may be subwords, including <unk>'s
            num_tokens += len(seq_word_ids)
            # number of words, including <s>'s and <unk>'s
            num_words += len(merged_words)
            # number of word probabilities computed (may not include <unk>'s)
            num_seq_probs = sum((lp is not None) and (not numpy.isneginf(lp))
                                for lp in merged_logprobs)
            num_probs += num_seq_probs
            # number of unks and zeroprobs (just for reporting)
            num_unks += sum(lp is None for lp in merged_logprobs)
            num_zeroprobs += sum((lp is not None) and numpy.isneginf(lp)
                                 for lp in merged_logprobs)
            # number of sequences
            num_sentences += 1

            if word_level:
                output_file.write("# Sentence {0}\n".format(num_sentences))
                _write_word_scores(vocabulary, merged_words, merged_logprobs,
                                   output_file, log_scale)
                output_file.write("Sentence perplexity: {0}\n\n".format(
                    numpy.exp(-seq_logprob / num_seq_probs)))

    output_file.write("Number of sentences: {0}\n".format(num_sentences))
    output_file.write("Number of words: {0}\n".format(num_words))
    output_file.write("Number of tokens: {0}\n".format(num_tokens))
    output_file.write("Number of predicted probabilities: {0}\n"
                      .format(num_probs))
    output_file.write("Number of excluded (OOV) words: {0}\n"
                      .format(num_unks))
    output_file.write("Number of zero probabilities: {0}\n"
                      .format(num_zeroprobs))
    if num_words > 0:
        cross_entropy = -total_logprob / num_probs
        perplexity = numpy.exp(cross_entropy)
        output_file.write("Cross entropy (base e): {0}\n".format(cross_entropy))
        if log_base is not None:
            cross_entropy /= log_scale
            output_file.write("Cross entropy (base {1}): {0}\n".format(
                cross_entropy, log_base))
        output_file.write("Perplexity: {0}\n".format(perplexity))

def _merge_subwords(subwords, subword_logprobs, marking):
    """Creates a word list from a subword list.

    :type subwords: list of strs
    :param subwords: list of vocabulary words, which may be subwords

    :type subword_logprobs: list of floats
    :param subword_logprobs: list of log probabilities for each word/subword
                             starting from the second one, containing ``None``
                             in place of any ignored <unk>'s

    :type marking: str
    :param marking: ``None`` for word vocabulary, otherwise the type of subword
                    marking used: "word-boundary" if a word boundary token (<w>)
                    is used and "prefix-affix" if subwords are prefixed/affixed
                    with + when they can be concatenated

    :rtype: tuple of two lists
    :returns: the first item is a list of words, where each word is either a
              string or a list of subword strings; the second item is a list of
              log probabilities for each word, starting from the second one,
              containing ``None`` in place of any ignored <unk>'s
    """

    if len(subword_logprobs) != len(subwords) - 1:
        raise ValueError("Number of logprobs should be exactly one less than "
                         "the number of words.")

    if marking is None:
        # Vocabulary is already words.
        return subwords, subword_logprobs

    words = [[subwords[0]]]
    logprobs = []
    current_word = []
    current_logprob = 0.0

    if marking == 'word-boundary':
        # Words are separated by <w>. Merge subwords and logprobs between <w>
        # tokens. If any part of a word is <unk>, the whole word is <unk>.
        for subword, logprob in zip(subwords[1:], subword_logprobs):
            if (current_logprob is None) or (logprob is None):
                current_logprob = None
            else:
                current_logprob += logprob
            if subword == '<w>':
                if current_word:
                    words.append(current_word)
                    logprobs.append(current_logprob)
                    current_word = []
                current_logprob = 0.0
            elif ('<unk>' in current_word) or (subword == '<unk>'):
                current_word = ['<unk>']
            else:
                current_word.append(subword)
    elif marking == 'prefix-affix':
        # Merge subword to the current word if the current word ends in + and
        # the subword starts with +.
        for subword, logprob in zip(subwords[1:], subword_logprobs):
            if current_word and current_word[-1].endswith('+') and \
               subword.startswith('+'):
                current_word.append(subword)
                if (current_logprob is None) or (logprob is None):
                    current_logprob = None
                else:
                    current_logprob += logprob
            else:
                if current_word:
                    words.append(current_word)
                    logprobs.append(current_logprob)
                current_word = [subword]
                current_logprob = logprob
    else:
        raise ValueError("Invalid subword marking type: " + marking)

    if current_word:
        words.append(current_word)
        logprobs.append(current_logprob)
    return words, logprobs

def _write_word_scores(vocabulary, words, logprobs, output_file, log_scale):
    """Writes word-level scores to an output file.

    :type vocabulary: Vocabulary
    :param vocabulary: vocabulary for printing information about which word are
                       predicted by the network

    :type words: list
    :param words: either word strings or one list for each word that contains
                  the subwords

    :type logprobs: list of floats
    :param logprobs: logprob of each word starting from the second word

    :type output_file: file object
    :param output_file: a file where to write the output

    :type log_scale: float
    :param log_scale: divide logprobs by this amount to convert to correct base
    """

    def word_info(vocabulary, word):
        """Returns information whether a word is in the vocabulary and in the
        shortlist.
        """
        if not word in vocabulary:
            return word + ':OOV'
        word_id = vocabulary.word_to_id[word]
        if vocabulary.in_shortlist(word_id):
            return word + ':shortlist'
        else:
            return word + ':OOS'

    if len(logprobs) != len(words) - 1:
        raise ValueError("Number of logprobs should be exactly one less than "
                         "the number of words.")

    logprobs = [None if x is None else x / log_scale for x in logprobs]
    for index, logprob in enumerate(logprobs):
        if index - 2 > 0:
            history_list = ['...']
            history_list.extend(words[index - 2:index + 1])
        else:
            history_list = words[:index + 1]
        history_list = [' '.join(word) if isinstance(word, list) else word
                        for word in history_list]
        history = ' '.join(history_list)

        predicted = words[index + 1]
        if isinstance(predicted, list):
            info = ' '.join(word_info(vocabulary, subword)
                            for subword in predicted)
            predicted = ' '.join(predicted)
        else:
            info = word_info(vocabulary, predicted)

        if logprob is None:
            output_file.write("p({0} | {1}) is not predicted  [{2}]\n".format(
                predicted, history, info))
        else:
            output_file.write("log(p({0} | {1})) = {2}  [{3}]\n".format(
                predicted, history, logprob, info))

def _score_utterances(input_file, vocabulary, scorer, output_file,
                      log_base=None):
    """Reads utterances from ``input_file``, computes LM scores using
    ``scorer``, and writes one score per line to ``output_file``.

    Start-of-sentence and end-of-sentece tags (``<s>`` and ``</s>``) will be
    inserted at the beginning and the end of each utterance, if they're missing.
    Empty lines will be ignored, instead of interpreting them as the empty
    sentence ``<s> </s>``.

    :type input_file: file object
    :param input_file: a file that contains the input sentences in SRILM n-best
                       format

    :type vocabulary: Vocabulary
    :param vocabulary: vocabulary that provides mapping between words and word
                       IDs

    :type scorer: TextScorer
    :param scorer: a text scorer for rescoring the input sentences

    :type output_file: file object
    :param output_file: a file where to write the output n-best list in SRILM
                        format

    :type log_base: int
    :param log_base: if set to other than None, convert log probabilities to
                     this base
    """

    log_scale = 1.0 if log_base is None else numpy.log(log_base)

    for line_num, line in enumerate(input_file):
        lm_score = scorer.score_line(line, vocabulary)
        if lm_score is None:
            continue
        lm_score /= log_scale
        output_file.write(str(lm_score) + '\n')
        if (line_num + 1) % 1000 == 0:
            logging.info("%d sentences scored.", line_num + 1)

    if scorer.num_words == 0:
        logging.info("The input file contains no words.")
    else:
        logging.info("%d words processed, including start-of-sentence and "
              "end-of-sentence tags, and %d (%.1f %%) out-of-vocabulary "
              "words".format(scorer.num_words,
                             scorer.num_unks,
                             scorer.num_unks / scorer.num_words))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""A module that implements the Architecture class, a neural network
architecture description.
"""

import os

from theanolm.backend import IncompatibleStateError, InputError

class Architecture(object):
    """Neural Network Architecture Description

    A description of the neural network architecture can be read from a text
    file or from a neural network state stored in an HDF5 file.
    """

    def __init__(self, inputs, layers, output_layer=None):
        """Constructs a description of the neural network architecture.

        :type inputs: list of dict
        :param inputs: parameters for each input as a list of dictionaries

        :type layers: list of dict
        :param layers: parameters for each layer as a list of dictionaries

        :type output_layer: str
        :param output_layer: name of the layer that gives the output of the
                             network (the last layer if None)
        """

        self.inputs = inputs
        if not inputs:
            raise ValueError("Cannot construct Architecture without inputs.")

        self.layers = layers
        if not layers:
            raise ValueError("Cannot construct Architecture without layers.")

        if output_layer is None:
            self.output_layer = layers[-1]['name']
        else:
            self.output_layer = output_layer

    @classmethod
    def from_state(cls, state):
        """Constructs a description of the network architecture stored in a
        state.

        :type state: hdf5.File
        :param state: HDF5 file that contains the architecture parameters
        """

        if 'architecture' not in state:
            raise IncompatibleStateError(
                "Architecture is missing from neural network state.")
        h5_arch = state['architecture']

        if 'inputs' not in h5_arch:
            raise IncompatibleStateError(
                "Architecture parameter 'inputs' is missing from neural "
                "network state.")
        h5_inputs = h5_arch['inputs']
        inputs = []
        for input_id in sorted(h5_inputs.keys()):
            h5_input = h5_inputs[input_id]
            inputs.append(cls._read_h5_dict(h5_input))

        if 'layers' not in h5_arch:
            raise IncompatibleStateError(
                "Architecture parameter 'layers' is missing from neural "
                "network state.")
        h5_layers = h5_arch['layers']
        layers = []
        layer_ids = [int(x) for x in h5_layers.keys()]
        for layer_id in sorted(layer_ids):
            h5_layer = h5_layers[str(layer_id)]
            layers.append(cls._read_h5_dict(h5_layer))

        if 'output_layer' not in h5_arch.attrs:
            raise IncompatibleStateError(
                "Architecture parameter 'output_layer' is missing from "
                "neural network state.")
        output_layer = h5_arch.attrs['output_layer']

        return cls(inputs, layers, output_layer)

    @classmethod
    def from_description(cls, description_file):
        """Reads a description of the network architecture from a text file.

        :type description_file: file or file-like object
        :param description_file: text file containing the description

        :rtype: Network.Architecture
        :returns: an object describing the network architecture
        """

        inputs = []
        layers = []

        for line in description_file:
            fields = line.split()
            if not fields or fields[0][0] == '#':
                continue

            if fields[0] == 'input':
                input_description = dict()
                for field in fields[1:]:
                    parts = field.split('=', 1)
                    if len(parts) != 2:
                        raise InputError(
                            "'field=value' expected but '{}' found in an input "
                            "description in '{}'."
                            .format(field, description_file.name))
                    variable, value = parts
                    input_description[variable] = value
                if 'type' not in input_description:
                    raise InputError(
                        "'type' is not given in an input description in '{}'."
                        .format(description_file.name))
                if 'name' not in input_description:
                    raise InputError(
                        "'name' is not given in an input description in '{}'."
                        .format(description_file.name))
                inputs.append(input_description)

            elif fields[0] == 'layer':
                layer_description = {'inputs': [], 'devices': []}
                for field in fields[1:]:
                    parts = field.split('=', 1)
                    if len(parts) != 2:
                        raise InputError(
                            "'field=value' expected but '{}' found in a layer "
                            "description in '{}'."
                            .format(field, description_file.name))
                    variable, value = parts
                    if variable == 'size':
                        layer_description[variable] = int(value)
                    elif variable == 'input':
                        layer_description['inputs'].append(value)
                    elif variable == 'device':
                        layer_description['devices'].append(value)
                    else:
                        layer_description[variable] = value
                if 'type' not in layer_description:
                    raise InputError(
                        "'type' is not given in a layer description in '{}'."
                        .format(description_file.name))
                if 'name' not in layer_description:
                    raise InputError(
                        "'name' is not given in a layer description in '{}'."
                        .format(description_file.name))
                if not layer_description['inputs']:
                    raise InputError(
                        "'input' is not given in a layer description in '{}'."
                        .format(description_file.name))
                layers.append(layer_description)

            else:
                raise InputError("Invalid element in architecture "
                                 "description: {}".format(fields[0]))

        if not inputs:
            raise InputError("Architecture description contains no inputs.")
        if not layers:
            raise InputError("Architecture description contains no layers.")

        return cls(inputs, layers)

    @classmethod
    def from_package(cls, name):
        """Reads network architecture from one of the files packaged with
        TheanoLM.

        :type name: str
        :param name: name of a standard architecture file (without directory or
                     file extension)

        :rtype: Network.Architecture
        :returns: an object describing the network architecture
        """

        file_dir = os.path.abspath(os.path.dirname(__file__))
        package_dir = os.path.dirname(file_dir)
        description_path = os.path.join(package_dir,
                                        'architectures',
                                        name + '.arch')

        with open(description_path, 'rt', encoding='utf-8') as description_file:
            return cls.from_description(description_file)

    def get_state(self, state):
        """Saves the architecture parameters in a HDF5 file.

        The variable values will be saved as attributes of HDF5 groups. A
        group will be created for each level of the hierarchy.

        :type state: h5py.File
        :param state: HDF5 file for storing the architecture parameters
        """

        h5_arch = state.require_group('architecture')

        h5_inputs = h5_arch.require_group('inputs')
        for input_id, input_value in enumerate(self.inputs):
            h5_input = h5_inputs.require_group(str(input_id))
            self._write_h5_dict(h5_input, input_value)

        h5_layers = h5_arch.require_group('layers')
        for layer_id, layer in enumerate(self.layers):
            h5_layer = h5_layers.require_group(str(layer_id))
            self._write_h5_dict(h5_layer, layer)

        h5_arch.attrs['output_layer'] = self.output_layer

    def check_state(self, state):
        """Checks that the architecture stored in a state matches this
        network architecture, and raises an ``IncompatibleStateError``
        if not.

        :type state: h5py.File
        :param state: HDF5 file that contains the architecture parameters
        """

        if 'architecture' not in state:
            raise IncompatibleStateError(
                "Architecture is missing from neural network state.")
        h5_arch = state['architecture']

        if 'inputs' not in h5_arch:
            raise IncompatibleStateError(
                "Architecture parameter 'inputs' is missing from neural "
                "network state.")
        h5_inputs = h5_arch['inputs']
        for input_id, input_value in enumerate(self.inputs):
            h5_input = h5_inputs[str(input_id)]
            self._check_h5_dict(h5_input, input_value)

        if 'layers' not in h5_arch:
            raise IncompatibleStateError(
                "Architecture parameter 'layers' is missing from neural "
                "network state.")
        h5_layers = h5_arch['layers']
        for layer_id, layer in enumerate(self.layers):
            h5_layer = h5_layers[str(layer_id)]
            self._check_h5_dict(h5_layer, layer)

        if 'output_layer' not in h5_arch.attrs:
            raise IncompatibleStateError(
                "Architecture parameter 'output_layer' is missing from "
                "neural network state.")
        h5_value = h5_arch.attrs['output_layer']
        if self.output_layer != h5_value:
            raise IncompatibleStateError(
                "Neural network state has output_layer={1}, while "
                "this architecture has output_layer={0}.".format(
                    self.output_layer, h5_value))

    @staticmethod
    def _read_h5_dict(h5_dict):
        """Reads a dictionary from a HDF5 file.

        The dictionary can store strings, which are represented as HDF5
        attributes, and lists of strings, which are represented as attributes of
        HDF5 subgroups.

        :type h5_dict: h5py.Group
        :param h5_dict: The HD5 group that stores a dictionary of strings in its
                        attributes
        """

        result = dict()
        for variable in h5_dict.attrs:
            result[variable] = h5_dict.attrs[variable]
        for variable in h5_dict:
            values = []
            h5_values = h5_dict[variable]
            for value_id in sorted(h5_values.attrs.keys()):
                values.append(h5_values.attrs[value_id])
            result[variable] = values
        return result

    @staticmethod
    def _write_h5_dict(h5_group, variables):
        """Writes a dictionary to a HDF5 file.

        The dictionary can store strings, which are represented as HDF5
        attributes, and lists of strings, which are represented as attributes of
        HDF5 subgroups.

        :type h5_group: h5py.Group
        :param h5_group: The HD5 group that will store a dictionary of strings
                         in its attributes

        :type variables: dict
        :param variables: a dictionary that may contain strings and lists of strings
        """

        for variable, values in variables.items():
            if isinstance(values, list):
                h5_values = h5_group.require_group(variable)
                for value_id, value in enumerate(values):
                    h5_values.attrs[str(value_id)] = value
            else:
                h5_group.attrs[variable] = values

    @staticmethod
    def _check_h5_dict(h5_group, variables):
        """Checks that a dictionary matches a HDF5 group and raises an
        ``IncompatibleStateError`` if not.

        The dictionary can store strings, which are represented as HDF5
        attributes, and lists of strings, which are represented as attributes of
        HDF5 subgroups.

        :type h5_group: h5py.Group
        :param h5_group: The HD5 group that will store a dictionary of strings
                         in its attributes

        :type variables: dict
        :param variables: a dictionary that may contain strings and lists of
                          strings
        """

        for variable, values in variables.items():
            if isinstance(values, list):
                # For backward compatibility. Remove at some point.
                if (variable == 'devices') and (not variable in h5_group):
                    if not values:
                        continue
                    else:
                        raise IncompatibleStateError(
                            "Neural network state does not specify devices, "
                            "while this architecture uses {}."
                            .format(', '.join(str(x) for x in values)))
                h5_values = h5_group[variable]
                for value_id, value in enumerate(values):
                    h5_value = h5_values.attrs[str(value_id)]
                    if value != h5_value:
                        raise IncompatibleStateError(
                            "Neural network state has {0}={2}, while this "
                            "architecture has {0}={1}."
                            .format(variable, value, h5_value))
            else:
                h5_value = h5_group.attrs[variable]
                if values != h5_value:
                    raise IncompatibleStateError(
                        "Neural network state has {0}={2}, while this "
                        "architecture has {0}={1}."
                        .format(variable, values, h5_value))
